/**
  *  \file client/screens/sweepscreen.cc
  *  \brief Sweep screen
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/screens/sweepscreen.h"
#include "ui/window.h"
#include "ui/widgets/focusiterator.h"
#include "ui/widgets/multilinestatic.h"
#include "client/widgets/setselect.h"
#include "util/nls.h"
#include "gfx/palette.h"
#include "game/struct.h"
#include "ui/layout.h"
#include "ui/widget-busy.h"
#include "game/sweep.h"
#include "ui/msgbox.h"
#include "cpluslib/format.h"
#include "client/widgets/imagepanel.h"
#include "ui/widgets/alignwidget.h"
#include "resmgr/resmgr.h"
#include "ui/widgets/button.h"
#include "ui/widgets/checkbox.h"
#include "client/widgets/helpwidget.h"
#include "ui/richhelp.h"
#include "ui/richdocument.h"

namespace {
    enum { WIDTH = 500 };

    /** Sweep screen window. */
    class WSweepWindow : public WImagePanel {
        ObjectHolder       h;
        WPlayerSetSelector set_sel;
        UICheckbox         cb_database;

        void init();
     public:
        WSweepWindow();
        void setSelectedPlayers(GPlayerSet ps)
            { set_sel.setSelectedPlayers(ps); }
        GPlayerSet getSelectedPlayers() const
            { return set_sel.getSelectedPlayers(); }
        void onEnter();
    };

    /** Sweep processor with "busy" display. This will entertain the user with
        a "busy" widget in case deletion takes some time. */
    class WSweepBusyProcessor : public GSweepRunProcessor,
                                public UIBusyWidget
    {
     public:
        void removeFile(string_t name)
            {
                GSweepRunProcessor::removeFile(name);
                UIBusyWidget::tick();
            }
    };
}

/** Constructor. */
WSweepWindow::WSweepWindow()
    : WImagePanel(_("Sweep"), requestPixmap(ResId("bg.sweep", ui_root->getExtent().w))),
      set_sel(1000, NUM_PLAYERS, 15*font_ems[FONT_NORMAL]),
      cb_database(1001, _("Delete database files?"), 'd', 2)
{
    init();
}

/** Build window content. */
void
WSweepWindow::init()
{
    // VBox
    //   UIMultilineStatic "Choose race:"
    //   HBox
    //     WPlayerSetSelector
    //     VBox
    //       UICheckbox "database"
    //       UISpacer
    //   UIMultilineStatic "blah..."
    //   HBox
    //     UIButton "OK"
    //     UIButton "Cancel"
    //     UISpacer
    //     UIButton "Help"

    set_sel.setSelectablePlayers(GPlayerSet::allUpTo(NUM_PLAYERS));

    UIGroup& group0  = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& group1  = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& group12 = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& group2  = h.add(new UIGroup(UIHBoxLayout::instance5));

    group12.add(cb_database);
    group12.add(h.add(new UISpacer()));

    group1.add(set_sel);
    group1.add(group12);

    UIButton& btn_ok     = h.add(new UIButton(1, _("OK"), SDLK_RETURN));
    UIButton& btn_cancel = h.add(new UIButton(2, _("Cancel"), SDLK_ESCAPE));

    btn_ok.sig_fired.add(this, &WSweepWindow::onEnter);
    btn_cancel.addStop();

    UIRichDocument& doc = h.add(new UIRichDocument(GfxPoint(WIDTH, 1), 0));
    doc.getDocument().setPageWidth(WIDTH);
    doc.getDocument().add(parseRichText(_("The <em>Sweep</em> function will remove player data files. "
                                          "By default, it will delete only files that can be "
                                          "restored by re-unpacking a result file. "
                                          "With the <em>Database</em> option, it will also delete "
                                          "the history database and configuration files.<br/>"
                                          "Select the files you want to remove using <kbd>Space</kbd> "
                                          "or using the mouse.")));
    doc.getDocument().finish();
    doc.adjustToDocumentSize();

    group2.add(btn_ok);
    group2.add(btn_cancel);
    group2.add(h.add(new UISpacer()));
    group2.add(h.add(new UIButton(_("Help"), 'h')));

    group0.add(h.add(new UIMultilineStatic(FONT_NORMAL, _("Choose player slots to clean:"), WIDTH)));
    group0.add(group1);
    group0.add(doc);
    group0.add(group2);
    group0.add(h.add(new UIFocusIterator(1000, 1001, UIFocusIterator::fi_Horiz | UIFocusIterator::fi_Tab)));

    add(h.add(new UIAlignedWidget(group0, GfxPoint(1, 1), GfxPoint(10, 10))));
    add(h.add(new WHelpWidget("pcc2:sweep")));

    setExtent(ui_root->getExtent());
    setState(st_Modal, true);
}

/** Handle "OK" button. This performs the actual sweep action. */
void
WSweepWindow::onEnter()
{
    /* Nothing selected? Just close this window. */
    if (set_sel.getSelectedPlayers().empty()) {
        stop(0);
        return;
    }

    /* Ask for confirmation */
    if (!UIMessageBox(_("Delete these players' files?"), _("Sweep")).doYesNoDialog()) {
        return;
    }

    /* Do it: */
    WSweepBusyProcessor runner;

    // ...configure sweep operation...
    GSweepControl control;
    control.setPlayers(set_sel.getSelectedPlayers());
    control.setEraseDatabase(cb_database.getValue());

    // ...display waitometer...
    ui_root->add(runner);
    ui_root->performDeferredRedraws();

    control.processFiles(runner);
    control.processIndex();

    // ...display result...
    messageBox(format(_("%d file%1{ has%|s have%} been deleted."), runner.getNumRemovedFiles()),
               _("Sweep"));

    /* Close window */
    stop(0);
}

/** Do Sweep screen. Asks user for parameters and deletes game files.
    \param defaultSet Default set to select for deletion. */
void
doSweep(GPlayerSet defaultSet)
{
    WSweepWindow window;
    window.setSelectedPlayers(defaultSet);
    ui_root->centerWidget(window);
    ui_root->runModal(window);
}
