/**
  *  \file client/screens/taskscreen.cc
  *  \brief Auto Task Screen
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/screens/taskscreen.h"
#include "client/tiles.h"
#include "int/exec.h"
#include "ui/msgbox.h"
#include "ui/widgets/inputline.h"
#include "util/nls.h"
#include "util/strlist.h"
#include "ui/listbox.h"
#include "ui/util.h"
#include "ui/dialogs/filedlg.h"
#include "io/textfile.h"
#include "util/except.h"
#include "cpluslib/format.h"

/** Constructor.
    \param base Underlying object type */
WAutoTaskObjectType::WAutoTaskObjectType(GObjectType& base)
    : GAliasObjectType(base),
      GObjectTypeObserver(base),
      base(base),
      current_process(0),
      conn_pos(base.sig_pos_changed.add(this, &WAutoTaskObjectType::onBaseObjectChanged)),
      cursor(0)
{
    sig_pos_changed.add(this, &WAutoTaskObjectType::onThisObjectChanged);
}

WAutoTaskObjectType::~WAutoTaskObjectType()
{
    done(); /* just in case */
}

/** Initialize. Call before using the auto task editor screen, but after
    finishing construction (virtual methods must be available). */
void
WAutoTaskObjectType::init()
{
    onThisObjectChanged();
}

/** Initialize. Call after using the auto task editor screen, but before
    starting destruction (virtual methods must be available). */
void
WAutoTaskObjectType::done()
{
    loadTask(0);
}

/** Set cursor. */
void
WAutoTaskObjectType::setCursor(uint32_t pos)
{
    if (pos != cursor) {
        cursor = pos;
        sig_task_changed.raise(false);
    }
}

/** Get cursor. */
uint32_t
WAutoTaskObjectType::getCursor() const
{
    return cursor;
}

/** Get editor. */
const IntAutoTaskEditor&
WAutoTaskObjectType::getEditor() const
{
    return editor;
}

/** Replace task text. \see IntAutoTaskEditor::replace */
void
WAutoTaskObjectType::replace(uint32_t pos, uint32_t nold, const string_t& line)
{
    replace(pos, nold, &line, 1);
}

/** Replace task text. \see IntAutoTaskEditor::replace */
void
WAutoTaskObjectType::replace(uint32_t pos, uint32_t nold, const string_t* line, uint32_t nnew)
{
    // Update task
    editor.replace(pos, nold, line, nnew);

    // Update cursor. We have the following cases:
    if (cursor < pos) {
        // cursor is before change; no modification
    } else if (cursor < pos+nold) {
        // cursor is within changed area: move to beginning of change
        cursor = pos;
    } else {
        // cursor is after changed area
        cursor -= nold;
        cursor += nnew;
    }

    // Update observers
    sig_task_changed.raise(false);
}

/** Insert one command. */
void
WAutoTaskObjectType::insert(const string_t& line)
{
    replace(getCursor(), 0, &line, 1);
}

/** Set program counter. */
void
WAutoTaskObjectType::setPC()
{
    editor.setPC(cursor);
    sig_task_changed.raise(false);
}

/** Common command GUI.
    \param sender Sender widget, for meun
    \param allowName Allow "SetName" command */
void
WAutoTaskObjectType::createCommonCommand(UIBaseWidget& sender, bool allowName)
{
    // Create list
    StringList sl;
    sl.add(1, _("Notify message"));
    sl.add(2, _("Set Comment"));
    if (allowName) {
        sl.add(3, _("Set Name"));
    }
    sl.add(4, _("Wait One Turn"));
    sl.add(5, _("Restart task from beginning"));

    // Menu
    UIStandardListbox list(0, sl.getCount(), getTextWidth(*fonts[FONT_NORMAL], sl) + 20, sl);
    if (!list.doStandardMenu(sender)) {
        return;
    }

    // Act
    string_t val;
    switch (list.getKey()) {
     case 1:
        if (UIInputLine::run(val, 255, _("Notify"), _("Enter message:"), 30, 0)) {
            insert("Notify " + quoteString(val));
        }
        break;
     case 2:
        if (UIInputLine::run(val, 255, _("Set Comment"), _("Enter new comment:"), 30, UIInputLine::ilf_GameChars)) {
            insert("SetComment " + quoteString(val));
        }
        break;
     case 3:
        if (UIInputLine::run(val, 20, _("Set Name"), _("Enter new name:"), 20, UIInputLine::ilf_GameChars)) {
            insert("SetName " + quoteString(val));
        }
        break;
     case 4:
        insert("WaitOneTurn");
        break;
     case 5:
        insert("Restart");
        break;
    }
}

/** Called when selected object changes on this object type.
    Reflect this change to the base object, and load new task. */
void
WAutoTaskObjectType::onThisObjectChanged()
{
    int id = getCurrent();
    if (base.getCurrent() != id) {
        base.setCurrent(id);
    }
    loadTask(getTask(true));
    sig_task_changed.raise(true);
}

/** Called when selected object changes on the base object object type.
    Reflect this change to the this type, and load new task. */
void
WAutoTaskObjectType::onBaseObjectChanged()
{
    int baseId = base.getCurrent();
    if (baseId != getCurrent()) {
        setCurrent(baseId);
        loadTask(getTask(true));
        sig_task_changed.raise(true);
    }
}

/** Called when the object we're observing changes. */
void
WAutoTaskObjectType::onObjectChanged()
{
    if (current_process != 0) {
        sig_task_changed.raise(false);
    }
}

/** Load auto task.
    \param process Auto task process. Can be 0 to unload current process. */
void
WAutoTaskObjectType::loadTask(IntExecutionContext* process)
{
    /* Quick out */
    if (process == current_process) {
        return;
    }

    /* Flush original process and make it runnable */
    IntExecutionContext* old_process = current_process;
    if (current_process != 0) {
        editor.saveToProcess(*current_process);
        current_process->setState(IntExecutionContext::psRunnable);
    }

    /* Load new process, and freeze it */
    current_process = process;
    if (current_process != 0) {
        editor.loadFromProcess(*current_process);
        current_process->setState(IntExecutionContext::psFrozen);
    } else {
        editor.clear();
    }

    /* Set cursor */
    cursor = editor.getNumInstructions();

    /* Run old process */
    if (old_process != 0 && old_process->getState() == IntExecutionContext::psRunnable) {
        old_process->run();
    }
}


/**************************** WAutoTaskScreen ****************************/

/** Auto task screen.
    \param type Object type
    \param screenId Screen identifier */
WAutoTaskScreen::WAutoTaskScreen(WAutoTaskObjectType& type, WScreenId::ScreenId screenId)
    : WControlScreen(type, screenId),
      type(type)
{ }

WAutoTaskScreen::~WAutoTaskScreen()
{ }

bool
WAutoTaskScreen::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        if (event.key.code == SDLK_DELETE) {
            // Delete
            type.replace(type.getCursor(), 1, 0, 0);
            return true;
        } else if (event.key.code == SDLK_INSERT) {
            // Insert
            string_t cmd;
            while (UIInputLine::run(cmd, 4000, _("Insert Command"), _("Command:"), 40, 0)) {
                if (!IntAutoTaskEditor::isValidCommand(cmd)) {
                    messageBox(_("This is not a valid auto task command."), _("Error"));
                } else {
                    type.replace(type.getCursor(), 0, cmd);
                    break;
                }
            }
            return true;
        } else if (event.key.code == ' ') {
            // Edit
            string_t cmd;
            if (type.getCursor() < type.getEditor().getNumInstructions()) {
                cmd = type.getEditor()[type.getCursor()];
            }
            while (UIInputLine::run(cmd, 4000, _("Edit Command"), _("Command:"), 40, 0)) {
                if (!IntAutoTaskEditor::isValidCommand(cmd)) {
                    messageBox(_("This is not a valid auto task command."), _("Error"));
                } else {
                    type.replace(type.getCursor(), 1, cmd);
                    break;
                }
            }
            return true;
        } else if (event.key.code == '%' || event.key.code == 'k') {
            // Comment in/out
            if (type.getCursor() < type.getEditor().getNumInstructions()) {
                string_t cmd = type.getEditor()[type.getCursor()];
                if (cmd.size() > 0 && cmd[0] == '%') {
                    cmd.erase(0, cmd.find_first_not_of(" %", 0));
                } else {
                    cmd.insert(0, "% ");
                }
                if (IntAutoTaskEditor::isValidCommand(cmd)) {
                    type.replace(type.getCursor(), 1, cmd);
                }
            }
            return true;
        } else if (event.key.code == ss_Ctrl + SDLK_DELETE) {
            // Delete all
            if (UIMessageBox(_("Delete this auto task?"), _("Auto Task")).doYesNoDialog()) {
                type.replace(0, type.getEditor().getNumInstructions(), 0, 0);
            }
            return true;
        } else if (event.key.code == ss_Ctrl + 'n') {
            // Set PC
            type.setPC();
            return true;
        } else if (event.key.code == ss_Ctrl + 'r') {
            // read file
            handleLoad();
            return true;
        } else if (event.key.code == ss_Ctrl + 's') {
            // save file
            handleSave();
            return true;
        }
    }
    return WControlScreen::handleEvent(event, second_pass);
}

/** "Save" command. */
void
WAutoTaskScreen::handleSave()
{
    // Ask for file
    UIFileChooserDialog fc(_("Save Task"), Directory::openDir(Directory::getWorkingDirectory()), "*.cct");
    ui_root->centerWidget(fc);
    if (!ui_root->runModal(fc)) {
        return;
    }

    // Save
    try {
        Ptr<Stream> out = Stream::openFile(fc.getResultFile(), Stream::C_CREATE);
        TextFile tf(*out);
        tf.setUtf8();
        for (uint32_t i = 0, e = type.getEditor().getNumInstructions(); i < e; ++i) {
            tf.writeLine(type.getEditor()[i]);
        }
        tf.sync();
    }
    catch (FileProblemException& e) {
        messageBox(format(_("Unable to create file %s: %s"), e.getFileName(), e.what()), _("Save Task"));
    }
    catch (std::exception& e) {
        messageBox(format(_("Unable to create file %s: %s"), fc.getResultFile(), e.what()), _("Save Task"));
    }
}

/** "Load" command. */
void
WAutoTaskScreen::handleLoad()
{
    // Ask for file
    UIFileChooserDialog fc(_("Load Task"), Directory::openDir(Directory::getWorkingDirectory()), "*.cct");
    ui_root->centerWidget(fc);
    if (!ui_root->runModal(fc)) {
        return;
    }

    // Save
    try {
        Ptr<Stream> out = Stream::openFile(fc.getResultFile(), Stream::C_READ);
        TextFile tf(*out);
        tf.setUtf8();
        string_t line;
        int numBadLines = 0;
        while (tf.getLine(line)) {
            if (IntAutoTaskEditor::isValidCommand(line)) {
                type.insert(line);
            } else {
                ++numBadLines;
            }
        }
        if (numBadLines != 0) {
            messageBox(format(_("This file contained %d line%1{ which is%|s which are%} not permitted in Auto Tasks. "
                                "%1{It has%|They have%} been skipped."), numBadLines),
                       _("Load Task"));
        }
    }
    catch (FileProblemException& e) {
        messageBox(format(_("Unable to open file %s: %s"), e.getFileName(), e.what()), _("Load Task"));
    }
    catch (std::exception& e) {
        messageBox(format(_("Unable to open file %s: %s"), fc.getResultFile(), e.what()), _("Load Task"));
    }
}
