/**
  *  \file game/actions/baseparts.cc
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/actions/baseparts.h"
#include "util/nls.h"

const int GStarbaseBuildPartsAction::INDEX_TABLE[] = {
    // same order as enum TTechLevel
    GStarbaseBuildPartsAction::FirstEngine,
    GStarbaseBuildPartsAction::FirstHull,
    GStarbaseBuildPartsAction::FirstBeam,
    GStarbaseBuildPartsAction::FirstLauncher,
};

/** \class GStarbaseBuildPartsAction
    \brief Building Starship Components

    This class allows building an arbitrary set of starship components.
    It automatically accounts for required tech updates.

    Orders are maintained in a relative fashion, i.e. we maintain a set
    of differences ("buy 3 Transwarps"), not target amounts ("make sure
    we have 8 Transwarps"). */

/** Constructor.
    \param planet Planet to work on */
GStarbaseBuildPartsAction::GStarbaseBuildPartsAction(GPlanet& planet)
    : GStarbaseBuildAction(planet),
      have_undo_information(false),
      num_tech_updates_needed(0),
      need_reg_update(false),
      conn_update(planet.sig_changed.add(this, &GStarbaseBuildPartsAction::updateFromPlanet))
{
    // FIXME: should be exception, not assert
    ASSERT(planet.isPlayed());
    ASSERT(planet.hasBase());

    std::fill_n(differences, int(TotalSlots), 0);
    updateFromPlanet();
}
    
GStarbaseBuildPartsAction::~GStarbaseBuildPartsAction()
{
    /* GStarbaseBuildAction will cancel the costs */
}

/** Copy base's storage into Item.
    \param item   [out] Receives storage summary
    \param planet [in] Starbase to copy from */
void
GStarbaseBuildPartsAction::copyFromPlanet(Item& item, const GPlanet& planet)
{
    for (int i = 0; i < NUM_HULLS_PER_PLAYER; ++i)
        item[FirstHull + i] = planet.getBaseHullStoreSlot(i+1);
    for (int i = 0; i < NUM_ENGINES; ++i)
        item[FirstEngine + i] = planet.getBaseEngineStore(i+1);
    for (int i = 0; i < NUM_BEAMS; ++i)
        item[FirstBeam + i] = planet.getBaseBeamStore(i+1);
    for (int i = 0; i < NUM_TORPS; ++i)
        item[FirstLauncher + i] = planet.getBaseLauncherStore(i+1);
}

/** Update after change on planet. */
void
GStarbaseBuildPartsAction::updateFromPlanet()
{
    if (!have_undo_information)
        copyFromPlanet(minimum, getPlanet());
    sig_changed.raise();
}

/** Handle tech update. Internal function, to be called from
    doChanges() only. Updates the num_tech_updates_needed
    and need_reg_update fields in addition to perform the action
    on the Actor.
    \param a     Actor to work on
    \param kind  Tech level to change
    \param level Required level */
void
GStarbaseBuildPartsAction::handleTechUpdate(GStarbaseBuildExecuter& a, TTechLevel kind, int level)
{
    if (level > getPlanet().getBaseTech(kind)) {
        ++num_tech_updates_needed;
        if (level > 6)
            need_reg_update = true;
        a.setTech(kind, level);
    }
}

/** Register undo information. Notifies this transaction of the state
    at the beginning of the turn, to allow sale of parts.
    \param old_planet State at beginning of the turn. */
void
GStarbaseBuildPartsAction::addUndoInformation(const GPlanet& old_planet)
{
    copyFromPlanet(minimum, old_planet);
    have_undo_information = true;
}

/** Get minimum possible amount. We allow sale of parts up to this amount.
    \param kind Kind of component to query
    \param slot Slot to query (1..NUM_HULLS_PER_PLAYER/NUM_ENGINES/etc.) */
int
GStarbaseBuildPartsAction::getMinimumParts(TTechLevel kind, int slot)
{
    /* check what we had at the beginning of the turn */
    int result = minimum[INDEX_TABLE[kind] + (slot-1)];

    /* do we need anything of that for the build order? */
    const GPlanet& p = getPlanet();
    if (p.getBaseBuildHull() != 0) {
        int occupied_slot = 0, required_amount = 0;
        const TBuildOrder& o = p.getBaseBuildOrder();
        switch (kind) {
         case tl_Hull:
            occupied_slot = p.getBaseBuildOrderHullSlot();
            required_amount = 1;
            break;
         case tl_Engine:
            occupied_slot = o.engine;
            required_amount = getHull(p.getBaseBuildHull()).getNumEngines();
            break;
         case tl_Beam:
            occupied_slot = o.beam_type;
            required_amount = o.beam_count;
            break;
         case tl_Torpedo:
            occupied_slot = o.torp_type;
            required_amount = o.torp_count;
            break;
        }
        if (occupied_slot == slot && required_amount > result)
            result = required_amount;
    }

    return result;
}

/** Get number of currently existing parts.
    \param kind Kind of component to query
    \param slot Slot to query (1..NUM_HULLS_PER_PLAYER/NUM_ENGINES/etc.) */
int
GStarbaseBuildPartsAction::getExistingParts(TTechLevel kind, int slot)
{
    switch (kind) {
     case tl_Hull:
        return getPlanet().getBaseHullStoreSlot(slot);
     case tl_Engine:
        return getPlanet().getBaseEngineStore(slot);
     case tl_Beam:
        return getPlanet().getBaseBeamStore(slot);
     case tl_Torpedo:
        return getPlanet().getBaseLauncherStore(slot);
     default:
        return 0;
    }
}

/** Get total parts after build.
    \param kind Kind of component to query
    \param slot Slot to query (1..NUM_HULLS_PER_PLAYER/NUM_ENGINES/etc.) */
int
GStarbaseBuildPartsAction::getTotalParts(TTechLevel kind, int slot)
{
    return getExistingParts(kind, slot)
        + differences[INDEX_TABLE[kind] + (slot-1)];
}

/** Add parts. This schedules the required amount for buy/sale.
    Note that this does not check how much we can pay for, only whether
    we can hold/sell that much.

    \param kind Kind of component
    \param slot Slot to query (1..NUM_HULLS_PER_PLAYER/NUM_ENGINES/etc.)
    \param amount Amount to buy (negative to sell)
    \param partial If true, accept partial buy/sell. If false, either
                   buy all or none
    \return number of items bought (negative if sold) */
int
GStarbaseBuildPartsAction::addParts(TTechLevel kind, int slot, int amount, bool partial)
{
    int change;
    if (amount == 0) {
        change = 0;
    } else if (amount > 0) {
        int have = getTotalParts(kind, slot);
        const int limit = 10000;
        int can_add = have > limit ? 0 : limit - have;
        change = amount > can_add ? can_add : amount;
    } else {
        int have = getTotalParts(kind, slot);
        int min = getMinimumParts(kind, slot);
        int can_remove = have < min ? 0 : have - min;
        change = -amount > can_remove ? -can_remove : amount;
    }

    if ((change != amount && !partial) || change == 0)
        /* we're ordered to build a full set, but can't; or we
           don't change anything */
        return 0;

    differences[INDEX_TABLE[kind] + (slot-1)] += change;
    sig_changed.raise();
    return change;
}

/** Perform all changes. Used in price computation and for actual buying. */
void
GStarbaseBuildPartsAction::perform(GStarbaseBuildExecuter& a)
{
    const GPlanet& p = getPlanet();

    num_tech_updates_needed = 0;
    need_reg_update = false;

    /* hulls */
    int tech_needed = p.getBaseTech(tl_Hull);
    for (int i = 0; i < NUM_HULLS_PER_PLAYER; ++i) {
        if (int d = differences[INDEX_TABLE[tl_Hull] + i]) {
            a.setHullSlot(i+1, p.getBaseHullStoreSlot(i+1) + d, 0);
            int level = getHull(getTruehull(p.getOwner(), i+1)).getTech();
            if (level > tech_needed)
                tech_needed = level;
        }
    }
    handleTechUpdate(a, tl_Hull, tech_needed);

    /* engines */
    tech_needed = p.getBaseTech(tl_Engine);
    for (int i = 0; i < NUM_ENGINES; ++i) {
        if (int d = differences[INDEX_TABLE[tl_Engine] + i]) {
            a.setEngine(i+1, p.getBaseEngineStore(i+1) + d, 0);
            int level = getEngine(i+1).getTech();
            if (level > tech_needed)
                tech_needed = level;
        }
    }
    handleTechUpdate(a, tl_Engine, tech_needed);
    
    /* beams */
    tech_needed = p.getBaseTech(tl_Beam);
    for (int i = 0; i < NUM_BEAMS; ++i) {
        if (int d = differences[INDEX_TABLE[tl_Beam] + i]) {
            a.setBeam(i+1, p.getBaseBeamStore(i+1) + d, 0);
            int level = getBeam(i+1).getTech();
            if (level > tech_needed)
                tech_needed = level;
        }
    }
    handleTechUpdate(a, tl_Beam, tech_needed);

    /* tubes */
    tech_needed = p.getBaseTech(tl_Torpedo);
    for (int i = 0; i < NUM_TORPS; ++i) {
        if (int d = differences[INDEX_TABLE[tl_Torpedo] + i]) {
            a.setLauncher(i+1, p.getBaseLauncherStore(i+1) + d, 0);
            int level = getTorp(i+1).getTech();
            if (level > tech_needed)
                tech_needed = level;
        }
    }
    handleTechUpdate(a, tl_Torpedo, tech_needed);
}

GError
GStarbaseBuildPartsAction::getError()
{
    if (need_reg_update)
        return GError(GError::ePerm, _("Registered-only tech upgrade requested"));
    return GError();
}
