/**
  *  \file game/classicvcr.cc
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <stdexcept>
#include "game/classicvcr.h"
#include "game/config.h"
#include "game/host.h"
#include "game/spec.h"
#include "game/pvcr.h"
#include "game/tvcr.h"
#include "util/nls.h"
#include "cpluslib/assert.h"
#include "io/streams.h"
#include "io/io.h"
#include "game/error.h"

static int
getBuildPointMass(const GClassicVcrEntry& entry, int side)
{
    // vcrplay.pas::PALMass
    const GClassicVcr& vcr = entry.getVcr();
    if (vcr(side).isPlanet()) {
        // planet
        return vcr(side).getMass() - 100;
    } else if ((!host.isPHost() || !config.PALIncludesESB(vcr(side).getOwner())) && vcr(side).getGuessedHull() != 0) {
        // ship, type known, and we have HOST or PHost where PAL does not include ESB
        return getHull(vcr(side).getGuessedHull()).getMass();
    } else {
        // ship, type unknown, or build points include ESB
        return vcr(side).getMass();
    }
}

/*! \class GClassicVcrEntry

    A GClassicVcrEntry is to a TVcr what GBeam is to TBeam: the high-level
    representation.

    For VCRs, we have a three-stage representation:
    - TRawVcr is a VCR as it appears in VCR.DAT files;
    - Since the VCR representation is ambiguous, mostly for
      PlanetsHaveTubes things, it is unpacked into a GClassicVcr.
      Much code uses these;
    - GClassicVcrEntry is the high-level representation for "one battle".

    Right now, there is no fine-grain accessor interface. */
GClassicVcrEntry::GClassicVcrEntry(const GClassicVcr& vcr)
    : status(0),
      type(vcr_Unknown),
      capabilities(0),
      vcr(vcr)
{ }

GClassicVcrEntry::~GClassicVcrEntry()
{ }

void
GClassicVcrEntry::setType(GClassicVcrType type, uint16_t capabilities)
{
    // rememver values
    this->type = type;
    this->capabilities = capabilities;

    // invalidate result
    this->status = 0;
}

GClassicVcrType
GClassicVcrEntry::getType() const
{
    return type;
}

uint16_t
GClassicVcrEntry::getCapabilities() const
{
    return capabilities;
}


int
GClassicVcrEntry::getNumObjects() const
{
    return 2;
}

GVcrObject&
GClassicVcrEntry::getObject(int slot, bool after)
{
    if (after) {
        return result(slot);
    } else {
        return vcr(slot);
    }
}

int
GClassicVcrEntry::getOutcome(int slot)
{
    // Make sure result is known
    if (this->status == 0) {
        ProgressMonitor pm;
        prepareResult(NeedQuickOutcome, pm);
    }

    // Process it
    if (this->status < 0) {
        // Error
        return 0;
    } else {
        if (slot == 0) {
            if (this->status & VcrPlayer::VCRS_LEFT_DESTROYED) {
                return -1;
            } else if (this->status & VcrPlayer::VCRS_LEFT_CAPTURED) {
                return vcr(1).getOwner();
            } else {
                return 0;
            }
        } else {
            if (this->status & VcrPlayer::VCRS_RIGHT_DESTROYED) {
                return -1;
            } else if (this->status & VcrPlayer::VCRS_RIGHT_CAPTURED) {
                return vcr(0).getOwner();
            } else {
                return 0;
            }
        }
    }
}

GVcrEntry::Playability
GClassicVcrEntry::getPlayability()
{
    if (this->status == 0) {
        // Result not known, perform check
        Ptr<VcrPlayer> player = createPlayer();
        player->setVisualizer(new VcrNullVisualizer());
        if (!player->setCapabilities(capabilities) || player->checkVcr(vcr)) {
            this->status = VcrPlayer::VCRS_INVALID;
        }
    }
    return this->status == VcrPlayer::VCRS_INVALID ? IsDamaged : IsPlayable;
}

void
GClassicVcrEntry::prepareResult(int /*resultLevel*/, ProgressMonitor& pm)
{
    if (this->status == 0) {
        // Result not known, so compute it
        Ptr<VcrPlayer> player = createPlayer();
        player->setVisualizer(new VcrNullVisualizer());
        GClassicVcr tmp(vcr);
        if (!player->setCapabilities(capabilities) || player->checkVcr(tmp)) {
            // Cannot be played
            this->status = VcrPlayer::VCRS_INVALID;
        } else {
            // Can be played
            player->playVcr(tmp, pm);
            player->doneVcr(result);
            this->status = player->getStatusWord();
        }
    }
}

string_t
GClassicVcrEntry::getAlgorithmName() const
{
    static const char*const names[] = {
        N_("Host"),
        N_("Unknown"),
        N_("Unknown PHost"),
        N_("PHost 2"),
        N_("PHost 3"),
        N_("PHost 4"),
        N_("NuHost"),
    };
    return _(names[type]);
}

bool
GClassicVcrEntry::isESBActive()
{
    // Simple cases
    if (!config.AllowEngineShieldBonus())
        return false;
    if (config.AllowESBonusAgainstPlanets())
        return true;

    // ESB is active, but not against planets. So find a planet.
    if (vcr(1).isPlanet())
        return false;
    return true;
}

/** Compute scores for this VCR.
    \param scores [in/out] Scores will be added here
    \param side [in] Side to compute information here */
void
GClassicVcrEntry::computeScores(int32_t scores[NumScores], const int side) const
{
    // true iff we destroyed the other one
    const bool isVictor = (side
                           ? (getStatus() & VcrPlayer::VCRS_LEFT_DESTROYED) != 0
                           : (getStatus() & VcrPlayer::VCRS_RIGHT_DESTROYED) != 0);
    // true iff we were not destroyed nor captured
    const bool didSurvive = (side
                             ? (getStatus() & (VcrPlayer::VCRS_RIGHT_DESTROYED | VcrPlayer::VCRS_RIGHT_CAPTURED)) == 0
                             : (getStatus() & (VcrPlayer::VCRS_LEFT_DESTROYED | VcrPlayer::VCRS_LEFT_CAPTURED)) == 0);
    const int opp = !side;

    const GClassicVcr& vcr = getVcr();
    const GClassicVcr& resultVcr = getResult();
    const int myRace = vcr(side).getOwner();

    if (host.isPHost()) {
        const int32_t damageDone = resultVcr(opp).getDamage() - vcr(opp).getDamage();
        const int32_t theirMass = std::max(getBuildPointMass(*this, opp), 1);
        const int32_t myMass = std::max(getBuildPointMass(*this, side), 1);

        // Build points. PHost gives different points for aggressor and opponent
        int32_t aggMP = damageDone * theirMass * config.PALAggressorPointsPer10KT(myRace);
        int32_t oppMP = damageDone * theirMass * config.PALOpponentPointsPer10KT(myRace);
        if (isVictor) {
            // We won
            aggMP += theirMass * 100 * config.PALAggressorKillPointsPer10KT(myRace);
            oppMP += theirMass * 100 * config.PALOpponentKillPointsPer10KT(myRace);
        } else if (!vcr(opp).isPlanet() && (getStatus() & (side ? VcrPlayer::VCRS_LEFT_CAPTURED : VcrPlayer::VCRS_RIGHT_CAPTURED)) != 0) {
            // We captured them
            int32_t crewKilled = vcr(opp).getCrew() - resultVcr(opp).getCrew();
            int32_t pts = crewKilled * config.PALShipCapturePer10Crew(myRace) * 100;
            aggMP += pts;
            oppMP += pts;
        }

        // Swap, so that aggMP is the bigger value
        if (oppMP > aggMP) {
            std::swap(aggMP, oppMP);
        }

        if (vcr(opp).isPlanet()) {
            // Scale planet points. Scale big*big, small*small
            int scale1 = config.PALCombatBaseScaling(myRace);
            int scale2 = config.PALCombatPlanetScaling(myRace);
            if (scale1 > scale2) {
                std::swap(scale1, scale2);
            }
            oppMP = (oppMP * scale1) / 100;
            aggMP = (aggMP * scale2) / 100;
        }
        scores[EarnedBuildMillipointsMin] += oppMP;
        scores[EarnedBuildMillipointsMax] += aggMP;

        // Experience
        if (didSurvive && config.NumExperienceLevels() > 0) {
            // FIXME: EPCombatBoostLevel / EPCombatBoostRate
            scores[EarnedExperience] += (damageDone * theirMass * config.EPCombatDamageScaling()) / (100 * myMass);
            if (isVictor) {
                scores[EarnedExperience] += theirMass * config.EPCombatKillScaling() / myMass;
            }
        }
    } else {
        // Build points only for destruction of other side, and only for ship/ship fight.
        if (isVictor && !vcr(1).isPlanet()) {
            const int32_t bmp = 1000 * ((getBuildPointMass(*this, opp) / 100) + 1);
            scores[EarnedBuildMillipointsMin] += bmp;
            scores[EarnedBuildMillipointsMax] += bmp;
        }

        // Experience.
        // Actually, HOST gives +20 for destroying a ship, +40 for capturing one,
        // but since players don't see experience, pretend there is none.
    }

    // Tons are common for all hosts
    if (isVictor) {
        scores[EarnedTonsDestroyed] += vcr(opp).getMass();
    }
}

/** Store result. Fetches the result from the specified
    VCR Player. The player must have called playVcr(), but
    not yet doneVcr(). */
void
GClassicVcrEntry::setResultFromPlayer(VcrPlayer& player)
{
    player.doneVcr(result);
    status = player.getStatusWord();
}

/** Create a VCR player that can play this VCR.
    \return player object */
Ptr<VcrPlayer>
GClassicVcrEntry::createPlayer()
{
    return createPlayerForType(type);
}

/** Create a VCR player that can play VCRs of the specified type.
    \return player object */
Ptr<VcrPlayer>
GClassicVcrEntry::createPlayerForType(GClassicVcrType type)
{
    // NOTE: caller must call setCapabilities!
    switch (type) {
     case vcr_Host:
        return new VcrPlayerTHost(false);
     case vcr_NuHost:
        return new VcrPlayerTHost(true);
     case vcr_PHost4:
        return new VcrPlayerPHost(true);
     case vcr_PHost3:
        return new VcrPlayerPHost(true);
     case vcr_PHost2:
        return new VcrPlayerPHost(false);
     default:
        throw std::runtime_error("Invalid VCR type");
    }
}

/************************** GClassicVcrDatabase **************************/

/** Check whether VCR record bears the PHost magic number. */
static inline bool
hasPHostMagic(const TRawVcr& vcr)
{
    /* some docs claim that there is another magic, 65261, but I have
       never seen that one in the wild. Actually, all docs from PHost
       1.1 up to 2.13 as well as 3.x and 4.x use 48879. */
    return ((vcr.random_seed + vcr.signature) & 0xFFFFU) == 48879;
}

/** Construct blank VCR database. */
GClassicVcrDatabase::GClassicVcrDatabase()
    : content()
{ }

/** Construct VCR database from file. */
GClassicVcrDatabase::GClassicVcrDatabase(Stream& file)
    : content()
{
    load(file);
}

/** Destructor. */
GClassicVcrDatabase::~GClassicVcrDatabase()
{ }

/** Load a file. */
void
GClassicVcrDatabase::load(Stream& file)
{
    // read count
    char    buffer[TRawVcr::size];
    file.readT(buffer, 2);

    uint16_t count        = getUint16(buffer);
    bool     maybePHost   = false;
    uint16_t firstTemp    = 0;
    uint16_t capabilities = 0;
    GClassicVcrType type  = vcr_Host;

    // read entries
    while (count--) {
        file.readT(buffer, sizeof(buffer));
        TRawVcr rawVcr;
        getStructure(buffer, rawVcr);

        if (content.empty()) {
            // this is the first VCR
            maybePHost = hasPHostMagic(rawVcr);
            firstTemp  = rawVcr._planet_temp_code;
        } else if (maybePHost && !count) {
            // this is the last VCR. Some things to note:
            // - it may be PHost 2's config battle. Discard it but memorize type.
            // - it may be the bogus battle from CORR (all but name zeroed)
            if (hasPHostMagic(rawVcr) && rawVcr.battle_type >= 2) {
                // PHost config battle. We assume that the PConfig is correct
                // and do not extract it from the file. Just discard the config
                // battle. Check the version field, though.
                if ((rawVcr._planet_temp_code & 255) == 2) {
                    // PHost 2
                    type = vcr_PHost2;
                } else {
                    // PHost 1 or something entirely different
                    type = vcr_UnknownPHost;
                }
                break;
            }
            if (rawVcr.objects[0].owner == 0 || rawVcr.objects[1].owner == 0) {
                // CORR
                break;
            }
        }

        // Remember it
        content.push_back(new GClassicVcrEntry(GClassicVcr(rawVcr)));
    }

    // If it hasn't been detected as PHost 2, it might be 3 or newer
    if (maybePHost && type == vcr_Host) {
        type = vcr_PHost3;
        if (firstTemp & vcrcap_Valid) {
            capabilities = firstTemp & ~vcrcap_Valid;
        }
        if (capabilities != 0) {
            type = vcr_PHost4;
        }
    }

    // If it still looks like Host, it might be NuHost VCRs unpacked by c2nu
    if (type == vcr_Host && !content.empty() && content.front()->getVcr().getSignature() == 0x554E) {
        type = vcr_NuHost;
    }

    // OK, now store type in all VCRs.
    for (ptr_vector<GClassicVcrEntry>::iterator i = content.begin(); i != content.end(); ++i) {
        (*i)->setType(type, capabilities);
    }
}

int
GClassicVcrDatabase::getNumBattles() const
{
    return content.size();
}

GClassicVcrEntry&
GClassicVcrDatabase::getBattle(int nr)
{
    ASSERT(nr >= 0 && nr < int(content.size()));
    return *content[nr];
}

/** Append new entry to VCR database.
    \return the entry */
GClassicVcrEntry&
GClassicVcrDatabase::addEntry(const GClassicVcr& vcr)
{
    GClassicVcrEntry* e = new GClassicVcrEntry(vcr);
    content.push_back(e);
    return *e;
}
