/**
  *  \file game/hullfunc.cc
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "cpluslib/format.h"
#include "game/hullfunc.h"
#include "game/spec.h"
#include "game/struct.h"
#include "util/console.h"
#include "util/filepars.h"
#include "util/misc.h"
#include "util/nls.h"
#include "game/config.h"

/*************************** GBasicHullFunction **************************/

/** Constructor.
    \param id     [in] Id under which host refers to this function
    \param name   [in] Name (identifier) */
GBasicHullFunction::GBasicHullFunction(int id, string_t name)
    : id(id), name(name),
      description(), explanation(),
      implied_device(-1)
{ }

GBasicHullFunction::~GBasicHullFunction()
{ }

/** Change name of function. */
void
GBasicHullFunction::setName(const string_t name)
{
    this->name = name;
}

/** Set short description of function. This is what we show to users,
    a short one-liner.
    \param desc New description */
void
GBasicHullFunction::setDescription(const string_t desc)
{
    this->description = desc;
}

/** Set explanation of function. This is the detailed explanation
    shown upon user request. It can contain multiple lines.
    \param explanation new explanation */
void
GBasicHullFunction::setExplanation(const string_t explanation)
{
    this->explanation = explanation;
}

/** Add to explanation. Adds a new line to the existing explanation.
    \param explanation line to add (may include '\\n' but doesn't have to)
    \see setExplanation */
void
GBasicHullFunction::addToExplanation(const string_t explanation)
{
    if (this->explanation.size() > 0 && this->explanation[this->explanation.size()-1] != '\n')
        this->explanation += '\n';
    this->explanation += explanation;
}

/** Set implied device. Each device can imply another one (usually a lesser
    version of it), meaning that a ship having both will perform only the
    better one, or, a ship having the better one can also do what the lesser
    one does.
    \param implied_device Number of implied device, see getId() */
void
GBasicHullFunction::setImpliedDevice(int implied_device)
{
    this->implied_device = implied_device;
}

/** Get device Id.
    \return hf_XXX */
int
GBasicHullFunction::getId() const
{
    return id;
}

/** Get device name.
    \return name, see setName() */
const string_t&
GBasicHullFunction::getName() const
{
    return name;
}

/** Get device description.
    \return description, see setDescription() */
const string_t&
GBasicHullFunction::getDescription() const
{
    if (description.size())
        return description;
    else
        return name;
}

/** Get device explanation.
    \return explanation, see setExplanation() */
const string_t&
GBasicHullFunction::getExplanation() const
{
    return explanation;
}

/** Get implied device.
    \return implied device Id, or -1
    \see setImpliedDevice */
int
GBasicHullFunction::getImpliedDevice() const
{
    return implied_device;
}

/************************ GHullFunction ************************/

/** Constructor. Construct a hull function object referring
    to a basic, unrestricted function. The function is assigned
    to all players, all levels.
    \param basic_function_id Basic function Id, see GBasicHullFunction::getId() */
GHullFunction::GHullFunction(int basic_function_id)
    : basic_function_id(basic_function_id),
      players(GPlayerSet::allUpTo(NUM_PLAYERS)),
      levels(GExpLevelSet::allUpTo(NUM_EXPERIENCE_LEVELS)),
      kind(AssignedToShip),
      host_id(-1)
{ }

/** Constructor. Construct a hull function object referring
    to a function restricted to a certain set of levels.
    \param basic_function_id Basic function Id, see GBasicHullFunction::getId()
    \param levels Levels at which this function works */
GHullFunction::GHullFunction(int basic_function_id, GExpLevelSet levels)
    : basic_function_id(basic_function_id),
      players(GPlayerSet::allUpTo(NUM_PLAYERS)),
      levels(levels),
      kind(AssignedToShip),
      host_id(-1)
{ }

/** Set player restriction.
    \param players Player set */
void
GHullFunction::setPlayers(GPlayerSet players)
{
    this->players = players;
}

/** Set level restriction.
    \param levels Level set */
void
GHullFunction::setLevels(GExpLevelSet levels)
{
    this->levels = levels;
}

/** Set kind. This describes how this function is assigned to a ship.
    \param kind Kind of assignment */
void
GHullFunction::setKind(Kind kind)
{
    this->kind = kind;
}

/** Set host Id. This is the number under which a level-restricted function
    is known to the host. */
void
GHullFunction::setHostId(int host_id)
{
    this->host_id = host_id;
}

/** Set basic function Id.
    \param basic_function_id Function Id, hf_XXX */
void
GHullFunction::setBasicFunctionId(int basic_function_id)
{
    this->basic_function_id = basic_function_id;
}

/** Get player restriction. */
GPlayerSet
GHullFunction::getPlayers() const
{
    return players;
}

/** Get level restriction. */
GExpLevelSet
GHullFunction::getLevels() const
{
    return levels;
}

/** Get kind of assignment. */
GHullFunction::Kind
GHullFunction::getKind() const
{
    return kind;
}

/** Get host Id.
    \return host id, or -1 if none */
int
GHullFunction::getHostId() const
{
    return host_id;
}

/** Get basic function Id. */
int
GHullFunction::getBasicFunctionId() const
{
    return basic_function_id;
}

/** Check whether two functions name the same hull function. This compares
    just the function data, not assignment information. */
bool
GHullFunction::isSame(const GHullFunction& other)
{
    return basic_function_id == other.basic_function_id
        && levels == other.levels;
}

/************************** GHullFunctionReader **************************/


namespace {
    /** Parser for 'hullfunc.cc'. */
    class GHullFunctionReader : public FileParser {
        /** Last seen hull function. Null if none yet. */
        GBasicHullFunction*  last_function;

        /** If set, the last hull function line was bogus, and we're
            ignoring text until the next one. */
        bool last_bogus;

        /** Reference to data store. */
        GHullFunctionData& data;
     public:
        GHullFunctionReader(GHullFunctionData& data)
            : FileParser(";#"), last_function(0), last_bogus(false), data(data)
            { }
        void process(int line_id, string_t line);
        void error(int line_id, string_t text);
    };
}

/** Shortcut function to display errors. */
void
GHullFunctionReader::error(int line_id, string_t text)
{
    console.write(LOG_ERROR,
                  format(_("%s:%d: %s"))
                  << getFileName() << line_id << text);
}

void
GHullFunctionReader::process(int line_id, string_t line)
{
    line = strTrim(line);

    string_t::size_type pos = line.find_first_of("=,");
    if (pos == string_t::npos) {
        error(line_id, _("Syntax error"));
        return;
    }

    if (line[pos] == ',') {
        /* it's a new function */
        string_t flags, name;
        string_t number = strRTrim(line.substr(0, pos));
        int      real_num;

        line.erase(0, pos+1);
        if (!strSplit(line, flags, name, ",")) {
            error(line_id, _("Syntax error"));
            return;
        }
        name = strTrim(name);

        if (!stringValue(number, real_num) || real_num < 0 || real_num > 0x7FFF) {
            last_bogus = true;
            error(line_id, format(_("Invalid device number for `%s'"), name));
            return;
        }

        /* avoid dupes */
        if (data.getBasicFunctionByName(name)) {
            last_bogus = true;
            error(line_id, format(_("Duplicate definition for hull function with name `%s'"), name));
            return;
        }
        if (data.getBasicFunctionById(real_num)) {
            last_bogus = true;
            error(line_id, format(_("Duplicate definition for hull function #%d"), real_num));
            return;
        }

        last_function = data.addBasicFunction(real_num, name);
        last_bogus = false;
    } else {
        /* it's an assignment */
        if (last_bogus)
            return;
        if (!last_function) {
            error(line_id, _("Expected function definition"));
            return;
        }

        string_t name  = strTrim(line.substr(0, pos-1));
        string_t value = strTrim(line.substr(pos+1));
        if (stringMatch("Implies", name)) {
            /* 'i' takes device number or name */
            int  number;
            trimComments(value);
            if (!stringValue(value, number)) {
                if (const GBasicHullFunction* hf = data.getBasicFunctionByName(value)) {
                    number = hf->getId();
                } else {
                    error(line_id, format(_("Unknown hull function `%s'"), value));
                    return;
                }
            }
            if (number != last_function->getId())
                last_function->setImpliedDevice(number);
        } else if (stringMatch("Description", name)) {
            /* 'd' takes string */
            last_function->setDescription(value);
        } else if (stringMatch("Explanation", name)) {
            /* 'e' takes string */
            last_function->addToExplanation(value);
        } else if (stringMatch("Standard", name)) {
            /* 's' takes list of hull numbers */
            trimComments(value);
            do {
                string_t lhs = strTrim(strFirst(value, ","));
                int num;
                if (! stringValue(lhs, num) || num <= 0 || num > MAX_HULLS) {
                    error(line_id, format(_("Invalid hull number `%s'"), lhs));
                    break;
                }
                data.addDefaultAssignment(num, last_function->getId());
            } while (strRemove(value, ","));
        } else {
            /* ignore */
        }
    }
}

/************************ GHullFunctionData ***********************/

/** Constructor. Make blank object. */
GHullFunctionData::GHullFunctionData()
{ }

GHullFunctionData::~GHullFunctionData()
{ }

/** Clear. Forget all stored contents. */
void
GHullFunctionData::clear()
{
    ptr_vector<GHullFunction>().swap(modified_functions);
    ptr_vector<GBasicHullFunction>().swap(basic_functions);
    default_assignments.clear();
}

/** Load basic function definitions. This initializes the object with
    generic specification data according to a game directory. */
void
GHullFunctionData::loadBasicFunctionDefinitions()
{
    clear();

    GHullFunctionReader rdr(*this);
    rdr.parseGameFile("hullfunc.usr");
    rdr.parseGameFile("hullfunc.cc");
}

/** Given a host-supplied device Id, return equivalent internal Id. */
GHullFunctionData::dev_t
GHullFunctionData::getIdFromHostId(int host_id) const
{
    for (ptr_vector<GHullFunction>::size_type i = 0; i != modified_functions.size(); ++i)
        if (host_id == modified_functions[i]->getHostId())
            return dev_t(i + MODIFIED_FUNCTION_BASE_ID);
    return dev_t(host_id);
}

/** Given a function definition, return equivalent internal Id. If
    there is not yet an internal Id for that definition, allocate one.
    If that definition reports a Host Id for an existing definition,
    register that one as well.
    \param def Function definition with basic_function_id, levels, and
               host_id fields valid
    \return internal Id */
GHullFunctionData::dev_t
GHullFunctionData::getIdFromFunction(const GHullFunction& def)
{
    if (def.getLevels() == GExpLevelSet::allUpTo(NUM_EXPERIENCE_LEVELS)) {
        // This is an unmodified function, hence its internal Id
        // is the same as its basic function Id.
        return dev_t(def.getBasicFunctionId());
    } else {
        // This is a modified function. Check whether we know it
        // already.
        for (ptr_vector<GHullFunction>::size_type i = 0; i != modified_functions.size(); ++i) {
            if (modified_functions[i]->isSame(def)) {
                // We know it
                if (def.getHostId() >= 0 && modified_functions[i]->getHostId() < 0)
                    modified_functions[i]->setHostId(def.getHostId());
                return dev_t(i + MODIFIED_FUNCTION_BASE_ID);
            }
        }

        // When we're here, we do not know the function yet. Add it.
        modified_functions.push_back(new GHullFunction(def));
        return dev_t(modified_functions.size() - 1 + MODIFIED_FUNCTION_BASE_ID);
    }
}

/** Return definition of a hull function.
    \param id   [in] Internal Id
    \param asgn [out] Definition of the function. This is always filled in; if the
                internal Id is invalid, it is filled with an invalid function
                definition. Note that it is possible to have valid function assignments
                that refer to a basic function whose definition we don't know, i.e.
                getBasicFunctionById(asgn.getBasicFunctionId()) == 0.
    \return true if valid internal Id */
bool
GHullFunctionData::getFunctionDefinition(dev_t id, GHullFunction& asgn) const
{
    if (int32_t(id) >= MODIFIED_FUNCTION_BASE_ID && unsigned(id - MODIFIED_FUNCTION_BASE_ID) < modified_functions.size()) {
        // It's a modified function
        asgn = *modified_functions[id - MODIFIED_FUNCTION_BASE_ID];
        return true;
    } else {
        // It's an unmodified function
        // FIXME: make validation stronger
        if (id >= 0) {
            asgn = GHullFunction(id);
            return true;
        } else {
            asgn = GHullFunction(-1);
            return false;
        }
    }
}

/** Get definition of a basic function.
    \param id [in] Identifier of basic function (hf_XXX)
    \return Definition, or null if none known
    \post return != 0 => return->getId() == id */
const GBasicHullFunction*
GHullFunctionData::getBasicFunctionById(int id) const
{
    for (ptr_vector<GBasicHullFunction>::const_iterator i = basic_functions.begin(); i != basic_functions.end(); ++i)
        if ((*i)->getId() == id)
            return *i;
    return 0;
}

/** Get definition of a basic function by name.
    \param name    [in] Name of function
    \param partial [in] true to accept partial matches (i.e. "cool" matches "CoolsTo50")
    \return Definition, or null if none matches */
const GBasicHullFunction*
GHullFunctionData::getBasicFunctionByName(string_t name, bool partial) const
{
    GBasicHullFunction* rv = 0;

    name = strLCase(name);
    for (ptr_vector<GBasicHullFunction>::const_iterator i = basic_functions.begin(); i != basic_functions.end(); ++i) {
        const string_t iname = strLCase((*i)->getName());
        if (iname == name) {
            rv = *i;
            break;
        }
        if (partial && iname.length() > name.length() && iname.compare(0, name.length(), name) == 0) {
            rv = *i;
            partial = false;
        }
    }
    return rv;
}

/** Add basic function definition.
    \param id    [in] Identifier of function (hf_XXX)
    \param name  [in] Name of function ("CoolsTo50")
    \return new definition
    \pre !getBasicFunctionById(id)
    \post getBasicFunctionById(id) == return */
GBasicHullFunction*
GHullFunctionData::addBasicFunction(int id, string_t name)
{
    return basic_functions.push_back(new GBasicHullFunction(id, name));
}

/** Check whether two basic hull function identifiers match.
    \param requested_id [in] The function we're looking for, e.g. hf_Cloak
    \param found_id     [in] Function found on the ship, e.g. hf_AdvancedCloak.
    \return true iff requested_id is equal to found_id or one of the
    functions it implies */
bool
GHullFunctionData::matchBasicFunction(int requested_id, int found_id) const
{
    while (found_id != requested_id) {
        const GBasicHullFunction* found_func = getBasicFunctionById(found_id);
        if (!found_func)
            /* this function does not exist */
            return false;

        found_id = found_func->getImpliedDevice();
        if (found_id < 0)
            /* this function doesn't imply anything */
            return false;
    }

    /* found match */
    return true;
}


/** Add a default assignment. This models the host's built-in defaults, and
    therefore cannot handle restricted functions.
    \param hull_id Hull that receives a function
    \param basic_function_id Basic function to assign */
void
GHullFunctionData::addDefaultAssignment(int hull_id, int basic_function_id)
{
    default_assignments.push_back(std::make_pair(hull_id, basic_function_id));
}

/** Perform default assignments on a set of hulls.
    \param hulls [in/out] Hulls
    \see addDefaultAssignment */
void
GHullFunctionData::performDefaultAssignments(ptr_vector<GHull>& hulls) const
{
    for (std::size_t i = 0; i < default_assignments.size(); ++i) {
        int hull = default_assignments[i].first;
        if (hull > 0 && unsigned(hull) <= hulls.size())
            hulls[hull-1]->changeSpecialFunction(dev_t(default_assignments[i].second),
                                                 GPlayerSet::allUpTo(NUM_PLAYERS),
                                                 GPlayerSet(),
                                                 true /* assign to hull */);
    }
}

/** Get default assignment for a function. Some hull functions can have a
    variable default assignment, depending upon configuration or other
    properties of a hull. In the host, the "Init=Default" statement will
    consult the current configuration, and set the functions accordingly.
    We want to be able to support configuration that changes on the fly
    without reloading hull functions (i.e. when the player configures
    AllowOneEngineTowing=Yes, all ships magically receive the 'Tow' ability).
    This function figures out the current variable default for a hull/device.

    \param device Device identifier
    \param hull   Hull object

    \return players who will have the function by default

    Note that the classic hull functions assigned to fixed hull numbers
    (i.e. "44-46 = Gravitonic") are not handled here.

    Note: all hull functions that can have a nonempty result here must be
    listed in GHull::clearSpecialFunctions(). */
GPlayerSet
GHullFunctionData::getDefaultAssignment(dev_t device, const GHull& hull) const
{
    /* only unmodified functions have a default assignment, so we can
       check using device = basic function Id here */
    if (int32_t(device) == hf_Tow) {
        /* if AllowOneEngineTowing enabled or ship has more than one engine,
           everyone can tow with it */
        if (config.AllowOneEngineTowing() || hull.getNumEngines() > 1)
            return GPlayerSet::allUpTo(NUM_PLAYERS);
        else
            return GPlayerSet();
    } else if (int32_t(device) == hf_Boarding) {
        /* Tholians and Privateers, if enabled */
        GPlayerSet result;
        if (config.AllowPrivateerTowCapture())
            result += getPlayersOfRace(5);
        if (config.AllowCrystalTowCapture())
            result += getPlayersOfRace(7);
        return result;
    } else if (int32_t(device) == hf_AntiCloakImmunity) {
        /* as configured */
        return getPlayersWhereEnabled(config.AntiCloakImmunity);
    } else if (int32_t(device) == hf_PlanetImmunity) {
        /* Rebels and Klingons, if enabled */
        /* This also applies to the SSD, but that's handled differently */
        GPlayerSet result;
        if (!config.PlanetsAttackKlingons())
            result += getPlayersOfRace(4);
        if (!config.PlanetsAttackRebels())
            result += getPlayersOfRace(10);
        return result;
    } else if (int32_t(device) == hf_FullWeaponry) {
        /* Feds, if enabled */
        if (config.AllowFedCombatBonus())
            return getPlayersOfRace(1);
        else
            return GPlayerSet();
    } else {
        /* nothing special */
        return GPlayerSet();
    }
}

/** Global instance of GHullFunctionData used for game. */
GHullFunctionData GHullFunctionData::instance;
