/**
  *  \file game/sel.h
  *  \brief Selection Layers
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_SEL_H
#define PCC_V2_GAME_SEL_H

#include "game/struct.h"
#include "cpluslib/types.h"
#include "cpluslib/string.h"
#include "cpluslib/sigslot.h"

class GUniverse;

/** A single selection layer. Stores one set of selection bits. */
class GSelection {
 public:
    enum {
        NUM_BITS  = NUM_SHIPS + NUM_PLANETS,
        NUM_WORDS = (NUM_BITS+31)/32,

        FIRST_SHIP_INDEX   = 0,
        FIRST_PLANET_INDEX = NUM_SHIPS
    };

    GSelection();

    void      clear();
    void      copyFrom(const GUniverse& univ);
    void      copyTo(GUniverse& univ);
    void      limitToExistingObjects(const GUniverse& univ);

    // Counts
    int       getNumShips() const;
    int       getNumPlanets() const;

    // Bitwise access
    bool      get(uint32_t index) const;
    void      set(uint32_t index, bool value);

    // Word-wise access
    uint32_t  operator[](uint32_t word_index) const;
    uint32_t& operator[](uint32_t word_index);
    static uint32_t getShipMask(uint32_t word_index);
    static uint32_t getPlanetMask(uint32_t word_index);

 private:
    uint32_t data[NUM_WORDS];
};

/** Multiple selection layers. This class provides a container for
    multiple selection management. It manages multiple GSelection
    objects. The current selection is always in the objects
    themselves, and is copied into a GSelection whenever the current
    layer changes, and at some places where the value is needed within
    the GMultiSelection. */
class GMultiSelection {
 public:
    enum {
        NUM_SELECTION_LAYERS = 8
    };

    /** Opcodes for compiled selection expressions. A selection
        expression is compiled to a RPN operator string. Note that
        these opcodes are fixed as they are (indirectly) part of
        process serialisation. */
    enum {
        opAnd        = '&',     ///< AND. Pop twice, push once.
        opOr         = '|',     ///< OR. Pop twice, push once.
        opXor        = '^',     ///< XOR. Pop twice, push once.
        opNot        = '!',     ///< NOT. Pop once, push once.
        opCurrent    = 'c',     ///< Current. Push once: value of current selection.
        opShip       = 's',     ///< Ship. Push once: true iff this is a ship.
        opPlanet     = 'p',     ///< Planet. Push once: true iff this is a planet.
        opFirstLayer = 'A',     ///< First layer. Other layers obtained by adding the layer number. Pushes value of that layer.
        opZero       = '0',     ///< Zero. Push once: false.
        opOne        = '1'      ///< One. Push once: true.
    };

    static int         getCurrentSelectionLayer();
    static void        setCurrentSelectionLayer(int layer);
    static void        executeSelectionExpression(int target_layer, string_t compiled_expression);
    static GSelection& getSelectionLayer(int layer);
    static bool        isValidSelectionLayer(int layer);

    static Signal0     sig_selection_changed;
};

class ModuleInitializer;
extern const ModuleInitializer mod_sel[];

#endif
