/**
  *  \file tvcr.cc
  *  \brief THost VCR Implementation
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include "game/tvcr.h"
#include "game/spec.h"
#include "game/config.h"
#include "game/formula.h"

/** \class VcrPlayerTHost
    \brief THost VCR Player

    This is the THost VCR player, started as a port from the PCC
    v1.0.13 player.

    It includes the speed optimisations (fast forward, psrandom
    tables), but is not yet completely tuned. Things that remain are
    - maybe widen the internally used data values (TVcr) or shorten
      those in #Status
    - maybe add some shortcuts for visualisation vs. simulation */

/* Benchmark results: `testvcr tests/vcr/vcr2.dat . 100', vcr2.dat from
   Manos-II / 35:

   + obfuscated seed advance:  2.350 user
   + 1-side instead of !side:  2.480 user
   + rand_table_type=uchar:    2.520 user
   + original:                 2.540 user
   + #undef CONF_STATIC_VCR:   2.950 user

   New results, `testvcr tests/vcr/vcr2.dat specs 1000':

   + starting with:            2.840
   + skip fireBeamsAtFighter:  2.760
   + fighter_status_type=u8:   2.740
   + 20110917:                 2.580
   + some inlines, rdivadd:    2.370
   + 20120612:                 2.210

   Removing VCR_STATIC, but further optimisations (in particular, store
   GVcrObject in Status and pass Status& around):
   + 20120616                  2.160

   Fast access paths (f_getDamage/f_getKill) seems not to get much. */

static const int RANDOM_SIZE = 119;

static const VcrPlayerTHost::rand_table_type rand_table_1_20[RANDOM_SIZE] = {
   	9,	8,	11,	8,	5,	5,
	9,	10,	15,	2,	10,	4,	14,
	18,	1,	14,	15,	17,	2,	4,
	10,	13,	16,	17,	11,	10,	14,
	7,	2,	8,	13,	13,	18,	6,
	13,	12,	6,	12,	6,	14,	4,
	1,	20,	16,	16,	2,	8,	10,
	18,	4,	20,	16,	17,	15,	6,
	19,	16,	14,	2,	15,	11,	6,
	9,	17,	15,	4,	3,	12,	16,
	19,	12,	18,	11,	13,	13,	8,
	3,	2,	15,	5,	12,	6,	10,
	6,	9,	16,	20,	19,	18,	17,
	11,	1,	4,	12,	7,	13,	15,
	5,	7,	12,	3,	3,	7,	14,
	10,	18,	13,	3,	16,	14,	4,
	13,	9,	14,	2,	9,	7,	4,	15
};

static const VcrPlayerTHost::rand_table_type rand_table_1_100[RANDOM_SIZE] = {
	42,	36,	54,	39,	23,	21,
	41,	45,	73,	5,	47,	14,	71,
	89,	2,	70,	76,	83,	5,	16,
	50,	64,	78,	87,	53,	47,	66,
	33,	5,	37,	63,	61,	88,	29,
	62,	58,	26,	61,	30,	67,	16,
	2,	98,	78,	81,	7,	37,	46,
	88,	15,	99,	77,	82,	75,	25,
	96,	79,	69,	5,	71,	54,	25,
	43,	87,	75,	17,	13,	58,	78,
	96,	57,	87,	52,	63,	64,	36,
	14,	5,	73,	23,	58,	29,	48,
	27,	43,	77,	99,	95,	88,	84,
	55,	2,	15,	57,	33,	61,	76,
	22,	31,	61,	11,	13,	31,	70,
	45,	92,	61,	11,	80,	71,	14,
	62,	44,	70,	4,	40,	32,	18,	74
};

static const VcrPlayerTHost::rand_table_type rand_table_1_17[RANDOM_SIZE] = {
	8,	7,	10,	7,	5,	4,
	7,	8,	13,	2,	8,	3,	12,
	15,	1,	12,	13,	14,	2,	3,
	9,	11,	13,	15,	9,	8,	12,
	6,	2,	7,	11,	11,	15,	5,
	11,	10,	5,	11,	6,	12,	3,
	1,	17,	14,	14,	2,	7,	8,
	15,	3,	17,	13,	14,	13,	5,
	16,	14,	12,	2,	12,	10,	5,
	8,	15,	13,	4,	3,	10,	13,
	16,	10,	15,	9,	11,	11,	7,
	3,	2,	13,	5,	10,	5,	9,
	5,	8,	13,	17,	16,	15,	14,
	10,	1,	3,	10,	6,	11,	13,
	4,	6,	11,	3,	3,	6,	12,
	8,	16,	11,	3,	14,	12,	3,
	11,	8,	12,	2,	7,	6,	4,	13
};

void
VcrPlayerTHost::Status::clear()
{
    for (int i = 0; i < VCR_MAX_BEAMS; ++i)
        beam_status[i] = 0;
    for (int i = 0; i < VCR_MAX_TORPS; ++i)
        torp_status[i] = 0;
    for (int i = 0; i < VCR_MAX_FTRS; ++i)
        ftr_x[i] = ftr_active[i] = 0;
    cur_x = 0;
    damage_limit = 100;
    num_fighters_out = 0;
    side = 0;
}

/** Random number between 1 and 20. The infamous VCR random number
    generator. Fetches the value from a table.

    For those who don't know it yet: VCR.EXE has a table with 119
    "random" numbers in the range [0,1000]. During a fight, these are
    repeatedly used and scaled to the requested range; the seed field
    in a VCR record is the initial index into the array (1 meaning the
    first entry). This implementation of the VCR uses three tables
    with pre-scaled values to avoid the costly rescale operation.
    Although VCR.EXE actually uses five ranges, it turns out that we
    can safely omit two of them by re-scaling the tests.

    Our seed runs from 1 to RANDOM_SIZE, although this spelling of the
    random number generator also allows setting it to zero (meaning
    the same as RANDOM_SIZE). */
VcrPlayerTHost::rand_type
VcrPlayerTHost::psrandom_1_20()
{
    /* The obfuscated form makes gcc generate better code at -O2
       (only two references to `seed', no in-memory increment) */
    return rand_table_1_20[(seed = (seed >= RANDOM_SIZE ? 1 : seed+1))-1];
}

/** Random number between 1 and 100. @see psrandom_1_20(). */
VcrPlayerTHost::rand_type
VcrPlayerTHost::psrandom_1_100()
{
    return rand_table_1_100[(seed = (seed >= RANDOM_SIZE ? 1 : seed+1))-1];
}

/** Random number between 1 and 17. @see psrandom_1_20(). */
VcrPlayerTHost::rand_type
VcrPlayerTHost::psrandom_1_17()
{
    return rand_table_1_17[(seed = (seed >= RANDOM_SIZE ? 1 : seed+1))-1];
}

/** Compute a/b+plus, using variable rounding.
    - nuflag=0: IEEE rounding (nearest or even)
    - nuflag=1: arithmetic rounding (nearest or up) */
int32_t
VcrPlayerTHost::rdivadd(int32_t a, int32_t b, int32_t plus)
{
    int32_t x = a / b + plus;
    int32_t r = a % b;
    if (r*2 + (x&1|nuflag) > b)
        ++x;
    return x;
}

/** Check if planet. \return true iff side (0 or 1) is a freighter. */
bool
VcrPlayerTHost::isFreighter(const Status& st)
{
    return st.obj.getBeamCount() == 0
        && st.obj.getTorpLauncherCount() == 0
        && st.obj.getBayCount() == 0;
}

/** Hit object.
    \param side     who is hit
    \param damage   `damage' (explosive) power of weapon
    \param kill     `kill' (x-ray) power of weapon */
void
VcrPlayerTHost::hit(Status& st, int damage, int kill)
{
    int shld = -rdivadd(80*damage, st.obj.getMass() + 1, 1 - st.obj.getShield());
    if (shld < 0) {
        st.obj.setShield(0);
        long l = rdivadd(-80L*shld, st.obj.getMass() + 1, st.obj.getDamage() + 1);
        if (l > 9999)
            l = 9999;
        st.obj.setDamage(l);
        shld = 0;
        if (st.obj.isPlanet() && damage > 1) {
            int beam = 10 - status[1].obj.getDamage() / 10;
            // bug emulation
            if (beam <= 0) {
                if (status[0].obj.getBeamType() > 0)
                    status[1].obj.setBeamType(status[0].obj.getBeamType());
            } else {
                if (status[1].obj.getBeamType() > beam)
                    status[1].obj.setBeamType(beam);
            }
        }
    }

    if (st.obj.getShield() == 0) {
        int crew = st.obj.getCrew();
        if (!st.obj.isPlanet()) {
            crew = - rdivadd(80*kill, st.obj.getMass() + 1, -crew);
            if (crew < 0)
                st.obj.setCrew(0);
            else
                st.obj.setCrew(crew);
        }
    }
    st.obj.setShield(shld);
}

/** \name Fighters */
//@{

/** Start a fighter. Attempts to launch a fighter.
    \pre Object has fighters. */
void
VcrPlayerTHost::launchFighter(Status& st)
{
    for (int i = 0; i < VCR_MAX_FTRS; ++i) {
        if (!st.ftr_active[i]) {
            st.obj.addFighters(-1);
            st.ftr_active[i] = 1;
            st.ftr_x[i]      = st.cur_x;
            ++st.num_fighters_out;
            vis->startFighter(st.side, i);
            setStatFighters(st.side, st.obj.getFighterCount());
            return;
        }
    }
}

/** Start fighter. Start fighter from specified side, if
    preconditions fulfilled. */
void
VcrPlayerTHost::launchFighters(Status& st)
{
    if (st.obj.getBayCount() > 0) {
        int n = psrandom_1_20();
        if (n <= st.obj.getBayCount() && st.obj.getFighterCount() > 0 && st.num_fighters_out < VCR_MAX_FTRS) {
            launchFighter(st);
        }
    }
}

/** Fighter fires. The fighter on track i fires at its opponent,
    if it's close enough. */
void
VcrPlayerTHost::fighterShoot(Status& st, Status& opp, int i)
{
    if (std::abs(st.ftr_x[i] - opp.cur_x) < 20) {
        hit(opp, 2, 2);
        vis->fireBeam(st.side, i, -1, 1, 2, 2);
    }
}

/** Kill fighter. The fighter on track i, specified side, is killed
    and removed from the game. */
void
VcrPlayerTHost::killFighter(Status& st, int i)
{
    vis->killFighter(st.side, i);
    st.ftr_active[i] = 0;
    --st.num_fighters_out;
}

/** Fighter stuff.

    - Fighter movement
    - Fighters fire at enemy
    - Fighter intercept */
void
VcrPlayerTHost::fighterStuff()
{
    /* This used to process all movements, then all firings. However, since all these
       processes are 100% deterministic, order does not matter and I have fused movement
       and firing, for a little bit of speed. */
    for (int i = 0; i < VCR_MAX_FTRS; ++i) {
        /* left side */
        if (status[0].ftr_active[i] == 1) {
            if (status[0].ftr_x[i] > status[1].cur_x + 10) {
                status[0].ftr_active[i] = 2;
                status[0].ftr_x[i] -= 4;
            } else {
                status[0].ftr_x[i] += 4;
                fighterShoot(status[0], status[1], i);
            }
        } else if (status[0].ftr_active[i] == 2) {
            if (status[0].ftr_x[i] < status[0].cur_x) {
                status[0].obj.addFighters(+1);
                vis->landFighter(0, i);
                status[0].ftr_active[i] = 0;
                --status[0].num_fighters_out;
            } else {
                status[0].ftr_x[i] -= 4;
            }
        }

        /* right side */
        if (status[1].ftr_active[i] == 1) {
            if (status[1].ftr_x[i] < status[0].cur_x - 10) {
                status[1].ftr_active[i] = 2;
                status[1].ftr_x[i] += 4;
            } else {
                status[1].ftr_x[i] -= 4;
                fighterShoot(status[1], status[0], i);
            }
        } else if (status[1].ftr_active[i] == 2) {
            if (status[1].ftr_x[i] > status[1].cur_x) {
                status[1].obj.addFighters(+1);
                vis->landFighter(1, i);
                status[1].ftr_active[i] = 0;
                --status[1].num_fighters_out;
            } else {
                status[1].ftr_x[i] += 4;
            }
        }
    }

    if (status[0].num_fighters_out > 0 && status[1].num_fighters_out > 0) {
        for (int i = 0; i < VCR_MAX_FTRS; ++i) {
            if (status[0].ftr_active[i]) {
                for (int j = 0; j < VCR_MAX_FTRS; ++j) {
                    if (status[1].ftr_active[j]) {
                        if (status[0].ftr_x[i] == status[1].ftr_x[j]) {
                            /* fighter intercept */
                            int n = psrandom_1_100();
                            if (status[0].ftr_active[i] == 0) {
                                /* Tim is the king! A dead fighter can
                                   still fire! Our visualisation can't
                                   handle that, so use the short way.
                                   See tests/vcr/deadfire.vcr for an
                                   example fight. */
                                if (n >= 50)
                                    killFighter(status[1], j);
                            } else {
                                /* regular fighter intercept code */
                                if (n < 50) {
                                    vis->fireBeam(1, j, i, 1, 2, 2);
                                    killFighter(status[0], i);
                                } else {
                                    vis->fireBeam(0, i, j, 1, 2, 2);
                                    killFighter(status[1], j);
                                }
                            }
                        }
                    }
                }
            }
        }
    }
}
//@}

/** \name Beam Weapons. */
//@{

/** Recharge beams of one side. Beams recharge randomly up to
    level 100. */
inline void
VcrPlayerTHost::rechargeBeams(Status& st)
{
    const int mx = st.obj.getBeamCount();
    for (int i = 0; i < mx; ++i) {
        register int j = st.beam_status[i];
        if (psrandom_1_100() > 50 && j < 100) {
            st.beam_status[i] = j+1;
            vis->updateBeam(st.side, i);
        }
    }
}

/** Fire beam. Hits the enemy with a beam of object from, beam number
    which. */
inline void
VcrPlayerTHost::fireBeam(Status& st, Status& opp, int which)
{
    int charge = st.beam_status[which];
    const GBeam& b = getBeam(st.obj.getBeamType());
    int da = rdivadd(charge * b.f_getDamage(), 100, 0);
    int ki = rdivadd(charge * b.f_getKill(),   100, 0);

    if (raceId(st.obj.getOwner()) == 5)
        ki *= 3;

    hit(opp, da, ki);
    vis->fireBeam(st.side, -1 - which, -1, 1, da, ki);
    st.beam_status[which] = 0;
    vis->updateBeam(st.side, which);
}

/** Fire beams from specified object. Fires all beams that can
    fire. */
void
VcrPlayerTHost::fireBeams(Status& st, Status& opp)
{
    for (int i = 0, n = st.obj.getBeamCount(); i < n; ++i) {
        int n = psrandom_1_20();
        if (n < 7 && st.beam_status[i] > 50)
            fireBeam(st, opp, i);
    }
}

/** Fire at fighters. Fires the specified beam at an enemy fighter,
    if possible. */
void
VcrPlayerTHost::fireAtFighter(Status& st, Status& opp, int beam)
{
    int min_dist = 600;
    int ftr_id   = -1;

    for (int i = 0; i < VCR_MAX_FTRS; ++i) {
        if (opp.ftr_active[i]) {
            int j = std::abs(opp.ftr_x[i] - st.cur_x);
            if (j < min_dist) {
                min_dist = j;
                ftr_id   = i;
            }
        }
    }

    if (ftr_id >= 0) {
        vis->fireBeam(st.side, -1 - beam, ftr_id, 1, 2, 2 /* FIXME */);
        killFighter(opp, ftr_id);
        st.beam_status[beam] = 0;
        vis->updateBeam(st.side, beam);
    }
}

/** Fire beams at fighter. Fire all beams that can fire at enemy
    fighters. */
void
VcrPlayerTHost::fireBeamsAtFighter(Status& st, Status& opp)
{
    if (!opp.num_fighters_out) {
        /* enemy has no fighters, so just advance the seed; saves some
           2..3% run time */
        seed = (seed + st.obj.getBeamCount()) % RANDOM_SIZE;
    } else {
        for (int i = 0, n = st.obj.getBeamCount(); i < n; ++i) {
            int n = psrandom_1_20();
            if (st.beam_status[i] > 40 && n < 5)
                fireAtFighter(st, opp, i);
        }
    }
}
//@}

/** \name Torpedoes. */
//@{

/** Fire torpedo. Launches one torpedo from the specified
    object. */
inline void
VcrPlayerTHost::fireTorp(Status& st, Status& opp, int launcher)
{
    register int n = psrandom_1_100();
    if (n >= 35) {
        const GTorpLauncher& t = getTorp(st.obj.getTorpType());
        hit(opp, 2*t.f_getDamage(), 2*t.f_getKill());
        setStatTorpHit(st.side);
        vis->fireTorp(st.side, n, launcher);
    } else {
        vis->fireTorp(st.side, -n, launcher);
    }
}

/** Fire torpedo. Fires all ready torpedoes from the specified
    object at the enemy. */
void
VcrPlayerTHost::fireTorpedoes(Status& st, Status& opp)
{
    for (int i = 0, max = st.obj.getTorpLauncherCount(); i < max; ++i) {
        if (st.obj.getTorpCount() > 0) {
            int n = psrandom_1_17();
            if (st.torp_status[i] > 40 || (st.torp_status[i] > 30 && n < st.obj.getTorpType())) {
                st.obj.addTorpedoes(-1);
                st.torp_status[i] = 0;
                vis->updateTorp(st.side, i);
                fireTorp(st, opp, i);
            }
            st.torp_status[i]++;
            vis->updateTorp(st.side, i);
        }
    }
}
//@}

/** \name Red Tape. */
//@{

/** Charge weapons. If the specified object has full shields,
    it also starts with fully-charged weapons. */
void
VcrPlayerTHost::preloadWeapons(Status& st)
{
    if (st.obj.getShield() == 100) {
        for (int i = 0; i < VCR_MAX_BEAMS; ++i) {
            st.torp_status[i] = 30;
            st.beam_status[i] = 100;
        }
    }
}

/** Init. Set up for playing. */
void
VcrPlayerTHost::initVcr(const GClassicVcr& avcr)
{
    GClassicVcr tmpVcr(avcr);
    if (checkVcr(tmpVcr)) {
        status_word = VCRS_INVALID;
        return;
    }

    /* okay, we'll be able to play it. */
    status_word = 0;
    seed = tmpVcr.getSeed() % RANDOM_SIZE;

    /* clear status */
    for (int i = 0; i < 2; ++i) {
        Status& st = status[i];
        st.clear();
        st.obj = tmpVcr(i);
        st.side = i;
        st.obj.setShield(std::max(0, std::min(st.obj.getShield(), 100 - st.obj.getDamage())));
    }

    status[0].cur_x = 30;       // ship
    if (!status[1].obj.isPlanet())
        status[1].cur_x = 610;
    else
        status[1].cur_x = 570;

    if (isFreighter(status[0]))
        status[0].obj.setShield(0);
    if (!status[1].obj.isPlanet()) {
        if (isFreighter(status[1]))
            status[1].obj.setShield(0);
    } else {
        if (status[1].obj.getCrew() <= 0) {
            status[1].obj.setShield(0);
        }
    }

    preloadWeapons(status[0]);
    preloadWeapons(status[1]);

    status[0].damage_limit = (raceId(status[0].obj.getOwner()) == 2) ? 151 : 100;
    status[1].damage_limit = (raceId(status[1].obj.getOwner()) == 2) ? 151 : 100;
    if (nuflag && status[1].obj.isPlanet()) {
        status[1].damage_limit = 100;
    }

    time = 0;
    initStat();
    vis->init();
}

/** Check one side of VCR. Fixes out-of-range values, and
    returns true iff there was such a fix needed. */
bool
VcrPlayerTHost::checkVcrSide(GClassicVcr& vcr, int side)
{
    bool err = false;
    GVcrObject& o = vcr(side);

    if (o.getOwner() <= 0 || o.getOwner() > 12)
        o.setOwner(12);

    if (o.getBeamType() < 0 || o.getBeamType() > NUM_BEAMS) {
        o.setBeamType(0);
        o.setBeamCount(0);
        err = true;
    }
    if (o.getTorpType() < 0 || o.getTorpType() > NUM_TORPS) {
        o.setTorpType(0);
        o.setTorpLauncherCount(0);
        err = true;
    }
    if (o.getBeamCount() > VCR_MAX_BEAMS) {
        o.setBeamCount(VCR_MAX_BEAMS);
        err = true;
    }

    if (o.getTorpLauncherCount() > VCR_MAX_TORPS) {
        o.setTorpLauncherCount(VCR_MAX_TORPS);
        err = true;
    }

    // unsure about this one, but it cannot hurt.
    if (vcr.getSeed() < 0 || vcr.getSeed() > RANDOM_SIZE) {
        vcr.setSeed(1);
        err = true;
    }

    return err;
}

/** Construct VCR player. */
VcrPlayerTHost::VcrPlayerTHost(bool nuflag)
{
    this->nuflag = nuflag;
}

/** Check VCR. */
bool
VcrPlayerTHost::checkVcr(GClassicVcr& vcr)
{
    bool l = checkVcrSide(vcr, 0);
    bool r = checkVcrSide(vcr, 1);
    return l || r;
}

/** Finish up VCR. Take back fighters, explode, compute status. */
void
VcrPlayerTHost::doneVcr(GClassicVcr& avcr)
{
    ASSERT(status_word != VCRS_INVALID);

    // FIXME: explode fighters when mothership explodes -- NOT! too
    // expensive in visualisation
    for (int i = 0; i < VCR_MAX_FTRS; ++i) {
        if (status[0].ftr_active[i]) {
            status[0].obj.addFighters(+1);
            vis->landFighter(0, i);
            status[0].ftr_active[i] = 0;
            --status[0].num_fighters_out;
        }
        if (status[1].ftr_active[i]) {
            status[1].obj.addFighters(+1);
            vis->landFighter(1, i);
            status[1].ftr_active[i] = 0;
            --status[1].num_fighters_out;
        }
    }

    status_word = 0;
    if (status[1].obj.isPlanet()) {
        /* FIXME: it seems that the ship can capture the planet even
           if it explodes in HOST. This affects the simulator. */
        if (status[0].obj.getDamage() >= 100 || status[0].obj.getCrew() <= 0)
            status_word |= VCRS_LEFT_DESTROYED;
        if (status[1].obj.getDamage() >= 100)
            status_word |= VCRS_RIGHT_DESTROYED;
    } else {
        if (status[0].obj.getDamage() >= status[0].damage_limit) {
            status_word |= VCRS_LEFT_DESTROYED;
        } else if (status[0].obj.getCrew() <= 0) {
            if (status[0].obj.getDamage() < status[1].damage_limit)
                status_word |= VCRS_LEFT_CAPTURED;
            else
                status_word |= VCRS_LEFT_DESTROYED;
        }
        if (status[1].obj.getDamage() >= status[1].damage_limit) {
            status_word |= VCRS_RIGHT_DESTROYED;
        } else if (status[1].obj.getCrew() <= 0) {
            if (status[1].obj.getDamage() < status[0].damage_limit)
                status_word |= VCRS_RIGHT_CAPTURED;
            else
                status_word |= VCRS_RIGHT_DESTROYED;
        }
    }

    if (status_word == 0)
        status_word = VCRS_TIMEOUT;

    if (status_word & VCRS_LEFT_DESTROYED)
        vis->killObject(0);
    if (status_word & VCRS_RIGHT_DESTROYED)
        vis->killObject(1);

    avcr(0) = status[0].obj;
    avcr(1) = status[1].obj;
}

bool
VcrPlayerTHost::setCapabilities(uint16 cap)
{
    return (cap == 0);
}

/** Play one cycle. */
bool
VcrPlayerTHost::playCycle()
{
    if (status_word != 0)
        return false;

    if (status[0].obj.getDamage() >= status[0].damage_limit || status[1].obj.getDamage() >= status[1].damage_limit)
        return false;
    if (status[0].obj.getCrew() <= 0 || status[1].obj.getCrew() <= 0)
        if (!status[1].obj.isPlanet())
            return false;
    if (time >= 2000)
        return false;

    ++time;

    /* Movement */
    int distance = status[1].cur_x - status[0].cur_x;
    if (distance > 30) {
        ++status[0].cur_x;
        --distance;
        if (!status[1].obj.isPlanet()) {
            --status[1].cur_x;
            --distance;
        }
    }

    /* Beams */
    if (distance < 200)
        fireBeams(status[0], status[1]);
    fireBeamsAtFighter(status[0], status[1]);
    fireBeamsAtFighter(status[1], status[0]);
    if (distance < 200)
        fireBeams(status[1], status[0]);

    /* Torpedoes */
    if (distance < 300) {
        fireTorpedoes(status[0], status[1]);
        fireTorpedoes(status[1], status[0]);
    }

    /* Fighters */
    launchFighters(status[0]);
    launchFighters(status[1]);
    if (status[0].num_fighters_out > 0 || status[1].num_fighters_out > 0)
        fighterStuff();

    /* Recharge beams */
    rechargeBeams(status[0]);
    rechargeBeams(status[1]);

    return true;
}

/** Fast forward. In torp/torp battles, nothing happens until ships are
    in torpedo range. So start there instead.
    \todo the same applies to freighters */
void
VcrPlayerTHost::playFastForward()
{
    /* not applicable if we already have a result, have played
       a tick, or are fighting a planet. */
    if (status_word || time || status[1].obj.isPlanet())
        return;

    /* not applicable for carriers */
    if (status[0].obj.getBayCount() || status[1].obj.getBayCount())
        return;

    if (status[0].obj.getShield() != 100 || status[1].obj.getShield() != 100)
        return;

    /* Start at distance 304 (to compensate off-by-one errors,
       300 should suffice ;-) */
    status[0].cur_x = 168;
    status[1].cur_x = 472;
    time = 138;

    seed = (seed + 138 * 2 * (status[0].obj.getBeamCount() + status[1].obj.getBeamCount())) % RANDOM_SIZE;
}
//@}

/** \name Accessors */
//@{
int
VcrPlayerTHost::getBeamStatus(int side, int id)
{
    return status[side].beam_status[id];
}

int
VcrPlayerTHost::getTorpStatus(int side, int id)
{
    /* scale [0,40] to [0,100] */
    return status[side].torp_status[id] * 5 / 2;
}

int
VcrPlayerTHost::getTorpCount(int side)
{
    if (status[side].obj.getTorpLauncherCount())
        return status[side].obj.getTorpCount();
    return 0;
}

int
VcrPlayerTHost::getFighterCount(int side)
{
    if (status[side].obj.getBayCount())
        return status[side].obj.getFighterCount();
    return 0;
}

int
VcrPlayerTHost::getShields(int side)
{
    return status[side].obj.getShield();
}

int
VcrPlayerTHost::getCrew(int side)
{
    return status[side].obj.getCrew();
}

int
VcrPlayerTHost::getDamage(int side)
{
    return status[side].obj.getDamage();
}

int
VcrPlayerTHost::getFighterX(int side, int id)
{
    return status[side].ftr_x[id];
}

int
VcrPlayerTHost::getFighterStatus(int side, int id)
{
    return status[side].ftr_active[id];
}

int
VcrPlayerTHost::getObjectX(int side)
{
    return status[side].cur_x;
}

long
VcrPlayerTHost::getDistance()
{
    return (status[1].cur_x - status[0].cur_x) * 100L;
}
//@}

VcrPlayerTHost::StatusToken::StatusToken(VcrPlayerTHost::time_type t)
    : VcrStatus(t)
{ }

/** Save status. The easy mindless way -- just save everything. */
VcrStatus*
VcrPlayerTHost::getStatusToken()
{
    StatusToken* t = new StatusToken(time);
    t->status[0]   = status[0];
    t->status[1]   = status[1];
    t->seed        = seed;
    t->status_word = status_word;
    t->nuflag      = nuflag;
    return t;
}

/** Restore status. */
void
VcrPlayerTHost::setStatus(VcrStatus& st)
{
    StatusToken* t = checked_cast<StatusToken*>(&st);
    status[0]      = t->status[0];
    status[1]      = t->status[1];
    seed           = t->seed;
    status_word    = t->status_word;
    time           = t->time;
    nuflag         = t->nuflag;
}
