/**
  *  \file game/utilread.cc
  *  \brief UTILx.DAT Reading
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/utilread.h"
#include "util/except.h"
#include "io/io.h"
#include "game/utildata.h"
#include "io/streams.h"

/****************************** GUtilReader ******************************/

/*! \class GUtilReader
    This encapsulates the logic for reading UTILx.DAT. Derived classes
    can override process() to actually process the records. */

/** UTILx.DAT reader for reading the specified stream. */
GUtilReader::GUtilReader()
{ }

GUtilReader::~GUtilReader()
{ }

/** Read UTILx.DAT. Reads the file, and calls process() for each record,
    in order. */
void
GUtilReader::read(Stream& s)
{
    char header[4];
    while ((s.read(header, sizeof(header))) == sizeof(header)) {
        std::size_t size = getUint16(header, 2);
        int         type = getInt16(header, 0);
        if (size) {
            Array<char> data (size);
            if (s.read(data.ptr(), size) != size) {
                failure(s);
                return;
            }
            if (!process(type, size, data.ptr())) {
                return;
            }
        } else {
            if (!process(type, size, 0)) {
                return;
            }
        }
    }
}

/** Check whether this is a valid UTILx.DAT. Valid files start with a
    valid control record. Optionally, return the timestamp in *ts. 
    This does not modify the file position (that is, it restores the
    position after reading) so it can safely be called before
    process() on the same stream.
    
    \returns true iff file seems valid.
    \pre stream is seekable.
    \throws never, except when the stream throws. */
bool
GUtilReader::check(Stream& s, GTimestamp* ts)
{
    Stream::fpos_t pos = s.getPos();
    char header[4];
    TUtil13ControlMinimal data;
    if (s.read(header, sizeof header) != sizeof header
        || getInt16(header, 0) != 13
        || getUint16(header, 2) < data.size
        || getUint16(header, 2) > 1024 /* arbitrary, but attempt to reject text files */
        || !getStructure(s, data))
    {
        s.seek(pos);
        return false;
    }
    s.seek(pos);
    if (ts)
        *ts = data.timestamp;
    return true;
}
