/**
  *  \file game/vcrobj.cc
  *  \brief VCR Object
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/vcrobj.h"
#include "game/struct.h"
#include "io/io.h"
#include "util/recode.h"
#include "game/config.h"
#include "game/spec.h"

/******************************** GVcrObject *******************************/

/** Default constructor. */
GVcrObject::GVcrObject()
    : name()
{
    pod.mass = 0;
    pod.shield = 0;
    pod.damage = 0;
    pod.crew = 0;
    pod.id = 0;
    pod.owner = 0;
    pod.race_or_zero = 0;
    pod.picture = 0;
    pod.hull_or_zero = 0;
    pod.beam_type = 0;
    pod.beam_count = 0;
    pod.torp_type = 0;
    pod.torp_count = 0;
    pod.torp_lcount = 0;
    pod.bay_count = 0;
    pod.fighter_count = 0;
    pod.experience_level = 0;
    pod.is_planet = false;
}

/** Construct from TRawVcr. This unpacks a TRawVcr from a classic VCR file.
    \param raw Data from file
    \param side Side to unpack, 0 or 1 */
GVcrObject::GVcrObject(const TRawVcr& raw, int side)
{
    unpack(raw, side);
}

/** Destructor. */
GVcrObject::~GVcrObject()
{ }

/** Unpack TRawVcr. This unpacks a TRawVcr from a classic VCR file.
    \param raw Data from file
    \param side Side to unpack, 0 or 1 */
void
GVcrObject::unpack(const TRawVcr& raw, int side)
{
    // Copy everything
    const TVcrObject& o  = raw.objects[side];
    pod.mass             = raw.mass[side];
    pod.shield           = raw.shield[side];
    pod.damage           = o.damage;
    pod.crew             = o.crew;
    pod.id               = o.id;
    pod.owner            = o.owner;
    pod.race_or_zero     = o.race_or_zero;
    pod.picture          = o.picture;
    pod.hull_or_zero     = o.hull_or_zero;
    pod.beam_type        = o.beam_type;
    pod.beam_count       = o.beam_count;
    pod.torp_type        = o.torp_type;
    pod.bay_count        = o.bay_count;
    pod.is_planet        = (side != 0 && raw.battle_type != 0);
    pod.experience_level = o.experience_level;
    name                 = convertGameToUtf8(getBasicString(raw.objects[side].name));

    // Handle torps/fighters
    if (o._torp_launcher_count != 0) {
        if (pod.is_planet && config.PlanetsHaveTubes()) {
            // It's a planet with tubes
            pod.torp_lcount   = (o._torp_launcher_count & 0xFF);
            pod.torp_count    = (o._torp_launcher_count >> 8) & 0xFF;
            pod.fighter_count = o._ammo;
        } else {
            // It has just torps
            pod.torp_lcount   = o._torp_launcher_count;
            pod.torp_count    = o._ammo;
            pod.fighter_count = 0;
        }
    } else {
        // No torps, but may have regular fighters
        pod.torp_lcount   = 0;
        pod.torp_count    = 0;
        pod.fighter_count = (o.bay_count != 0 ? o._ammo : 0);
    }

    // Silent fixes, avoid confusing display
    if (pod.beam_type == 0) {
        pod.beam_count = 0;
    }
    if (pod.torp_type == 0) {
        pod.torp_lcount = pod.torp_count = 0;
    }
}

/** Pack into classic TRawVcr.
    \param raw [out] Output will be produced here
    \param side [in] Side to pack
    \retval true packing successful
    \retval false data cannot be packed, %raw left incomplete */
bool
GVcrObject::pack(TRawVcr& raw, int side)
{
    TVcrObject& o = raw.objects[side];

    // Simple fields
    storeBasicString(o.name, convertUtf8ToGame(name));
    o.damage           = pod.damage;
    o.crew             = pod.crew;
    o.id               = pod.id;
    o.owner            = pod.owner;
    o.race_or_zero     = pod.race_or_zero;
    o.picture          = pod.picture;
    o.hull_or_zero     = pod.hull_or_zero;
    o.beam_type        = pod.beam_type;
    o.beam_count       = pod.beam_count;
    o.experience_level = pod.experience_level;
    o.bay_count        = pod.bay_count;
    raw.mass[side]     = pod.mass;
    raw.shield[side]   = pod.shield;

    // Secondary weapons
    if (pod.torp_lcount != 0) {
        // It has torps
        o.torp_type = pod.torp_type;
        if (pod.is_planet && config.PlanetsHaveTubes()) { // FIXME: als Parameter
            // Planet with torps and fighters
            if (pod.torp_lcount > 255 || pod.torp_count > 255) {
                return false;
            }
            o._torp_launcher_count = 256*pod.torp_count + pod.torp_lcount;
            o._ammo                = pod.bay_count != 0 ? pod.fighter_count : 0;
        } else {
            // Normal torps
            if (pod.bay_count != 0) {
                return false;
            }
            o._torp_launcher_count = pod.torp_lcount;
            o._ammo                = pod.torp_count;
        }
    } else {
        // No torps
        o.torp_type            = 0;
        o._torp_launcher_count = 0;
        o._ammo                = pod.bay_count != 0 ? pod.fighter_count : 0;
    }
    return true;
}

/** Remember guessed hull. This will modify the object in place, so it's a bad
    idea to call this on VCRs you intend to export, but it'll save a few cycles
    in guessHull() later on. */
void
GVcrObject::setGuessedHull()
{
    setHull(getGuessedHull());
}

/** Check if this could be the specified hull.
    \param hull_id Hull to check for
    \return true iff this ship could be of the specified hull
    \pre isPlanet() */
bool
GVcrObject::canBeHull(int hull_id) const
{
    const GHull& hull = ::getHull(hull_id);
    if (getHull() != 0) {
        /* hull known - case closed */
        return (getHull() == hull_id);
    } else {
        // FIXME: handle THost's easter egg here? Nebulas with Transwarps are reported
        // with wrong picture
        // FIXME: check other hull properties?
        // - mass: VCR ship shouldn't be lighter than original - true?
        // - crew: can be smaller or larger

        /* picture must match */
        if (getPicture() != hull.getPicture())
            return false;
        /* must not have more beams/torps than hull allows */
        if (getBeamCount() > hull.getMaxBeams() || getTorpLauncherCount() > hull.getMaxLaunchers())
            return false;
        /* for fighter bays, the only k.o. criterion is that ship has
           fighters but hull has not. The number of bays can be
           smaller (NTP, damage) or larger (scotty bonus). */
        if (getBayCount() && !hull.getNumBays())
            return false;
        return true;
    }
}

/** Guess this ship's hull. Whereas getHull() returns the value from the combat
    record, which can be missing, this will attempt to guess the correct value.
    \return hull number, zero if ambiguous, impossible, or this is actually a planet */
int
GVcrObject::getGuessedHull() const
{
    /* planets don't have hulls. */
    if (isPlanet())
        return 0;

    /* see if PHost sent us the hull type. */
    if (int h = getHull())
        return h;

    /* otherwise, try all hulls. */
    int type = 0;
    for (int h = 1; h <= getNumHulls(); ++h) {
        if (canBeHull(h)) {
            if (type == 0)
                type = h;
            else
                return 0;       // ambiguous, can't guess
        }
    }
    return type;
}

/** Get ship picture. Whereas getPicture() just returns the value from the combat record,
    this will attempt to resolve the record back into a ship type, and give that ship's
    picture. Thus, it will reflect users' changes. */
int
GVcrObject::getGuessedShipPicture() const
{
    // FIXME: actually, this currently is a little pointless, as guessHull() uses
    // the same getPicture() as this one. PCC 1.x had two different picture Ids
    // in a hullspec entry, one for resolving ship types, one for display.
    if (isPlanet()) {
        return 0;
    } else if (int g = getGuessedHull()) {
        return ::getHull(g).getPicture();
    } else {
        return getPicture();
    }
}
