/**
  *  \file reshack/fontsynth.cc
  *  \brief Resource Editor: Synthesizing Characters
  *
  *  Some characters can be synthesized which saves a lot of work
  *  while assembling Unicode fonts for PCC2.
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "reshack/fontsynth.h"
#include "reshack/painter.h"
#include "reshack/colorsel.h"
#include "gfx/bmfont.h"

/** Synthesize frame character. This can be used for U+2500 .. U+257F.
    The shape is described by an octal number:
    - 0Nxxxx = number of segments for dashed lines, 0 for solid lines
    - 0xNxxx = shape of "up" wing, 0=none, 1=light, 2=heavy, 3=double
    - 0xxNxx = shape of "right" wing
    - 0xxxNx = shape of "down" wing
    - 0xxxxN = shape of "left" wing
    \param painter Draw character here
    \param shape Shape description */
void
synthesizeFrameCharacter(RHPainter& painter, uint16_t shape)
{
    // Figure out center
    GfxPoint size = painter.getSize();
    int xc = (size.x) / 2;
    int yc = (size.y - 1) / 2;

    // Clear it
    Ptr<GfxPixmap> pix = painter.getPixmap();
    drawSolidBar(*pix, GfxRect(GfxPoint(0, 0), size), FC_BLACK);

    // Thick lines for digit 2 or 3
    if ((shape & 07000) > 01000) {
        pix->drawVLine(GfxPoint(xc-1, 0), yc+1, FC_WHITE, SOLID_LINE, GFX_ALPHA_OPAQUE);
        pix->drawVLine(GfxPoint(xc+1, 0), yc+1, FC_WHITE, SOLID_LINE, GFX_ALPHA_OPAQUE);
    }
    if ((shape & 00700) > 00100) {
        pix->drawHLine(GfxPoint(xc, yc-1), size.x-xc, FC_WHITE, SOLID_LINE, GFX_ALPHA_OPAQUE);
        pix->drawHLine(GfxPoint(xc, yc+1), size.x-xc, FC_WHITE, SOLID_LINE, GFX_ALPHA_OPAQUE);
    }
    if ((shape & 00070) > 00010) {
        pix->drawVLine(GfxPoint(xc-1, yc), size.y-yc, FC_WHITE, SOLID_LINE, GFX_ALPHA_OPAQUE);
        pix->drawVLine(GfxPoint(xc+1, yc), size.y-yc, FC_WHITE, SOLID_LINE, GFX_ALPHA_OPAQUE);
    }
    if ((shape & 00007) > 00001) {
        pix->drawHLine(GfxPoint(0, yc-1), xc+1, FC_WHITE, SOLID_LINE, GFX_ALPHA_OPAQUE);
        pix->drawHLine(GfxPoint(0, yc+1), xc+1, FC_WHITE, SOLID_LINE, GFX_ALPHA_OPAQUE);
    }

    // Re-blank centers for digit 3
    if ((shape & 07000) == 03000)
        pix->drawVLine(GfxPoint(xc, 0), yc+1, FC_BLACK, SOLID_LINE, GFX_ALPHA_OPAQUE);
    if ((shape & 00700) == 00300)
        pix->drawHLine(GfxPoint(xc, yc), size.x-xc, FC_BLACK, SOLID_LINE, GFX_ALPHA_OPAQUE);
    if ((shape & 00070) == 00030)
        pix->drawVLine(GfxPoint(xc, yc), size.y-yc, FC_BLACK, SOLID_LINE, GFX_ALPHA_OPAQUE);
    if ((shape & 00007) == 00003)
        pix->drawHLine(GfxPoint(0, yc), xc+1, FC_BLACK, SOLID_LINE, GFX_ALPHA_OPAQUE);

    // Draw center for digit 1 and 2
    if ((shape & 07000) != 03000 && (shape & 07000) != 0)
        pix->drawVLine(GfxPoint(xc, 0), yc+1, FC_WHITE, SOLID_LINE, GFX_ALPHA_OPAQUE);
    if ((shape & 00700) != 00300 && (shape & 00700) != 0)
        pix->drawHLine(GfxPoint(xc, yc), size.x-xc, FC_WHITE, SOLID_LINE, GFX_ALPHA_OPAQUE);
    if ((shape & 00070) != 00030 && (shape & 00070) != 0)
        pix->drawVLine(GfxPoint(xc, yc), size.y-yc, FC_WHITE, SOLID_LINE, GFX_ALPHA_OPAQUE);
    if ((shape & 00007) != 00003 && (shape & 00007) != 0)
        pix->drawHLine(GfxPoint(0, yc), xc+1, FC_WHITE, SOLID_LINE, GFX_ALPHA_OPAQUE);

    // Gaps for dashed lines
    int ngaps = (shape & 070000) >> 12;
    if (ngaps != 0) {
        int npieces = 2*ngaps - 1;
        if ((shape & 07070) != 0) {
            // We have vertical lines
            for (int i = 1; i < npieces; i += 2) {
                int y1 = size.y * i / npieces;
                int y2 = size.y * (i+1) / npieces;
                drawSolidBar(*pix, GfxRect(0, y1, size.x, y2-y1), FC_BLACK);
            }
        } else {
            // We have horizontal lines
            for (int i = 1; i < npieces; i += 2) {
                int x1 = size.x * i / npieces;
                int x2 = size.x * (i+1) / npieces;
                drawSolidBar(*pix, GfxRect(x1, 0, x2-x1, size.y), FC_BLACK);
            }
        }
    }
}

/** Synthesize block character. This can be used for U+2580 .. U+259FF.
    The shape is described by an octal number:
    - 0Nxx: kind, 1=vertical split, 2=horizontal split, 3=quadrant, 4=shade
    - 0xNx: first cell
    - 0xxN: number of cells
    For quadrant and shade, the last two digits give four bits of pattern
    that defines which quadrants / pixels are drawn.
    \param painter Draw character here
    \param shape Shape description */
static void
synthesizeBlockCharacter(RHPainter& painter, uint16_t shape)
{
    // Figure out center
    GfxPoint size = painter.getSize();

    // Clear it
    Ptr<GfxPixmap> pix = painter.getPixmap();
    drawSolidBar(*pix, GfxRect(GfxPoint(0, 0), size), FC_BLACK);

    switch ((shape & 0700) >> 6) {
     case 1:
        // Vertical
        {
            int a = (shape & 070) >> 3;
            int b = (shape & 7);
            int y1 = size.y * a / 8;
            int y2 = size.y * (a+b+1) / 8;
            drawSolidBar(*pix, GfxRect(0, y1, size.x, y2-y1), FC_WHITE);
        }
        break;
     case 2:
        // Horizontal
        {
            int a = (shape & 070) >> 3;
            int b = (shape & 7);
            int x1 = size.x * a / 8;
            int x2 = size.x * (a+b+1) / 8;
            drawSolidBar(*pix, GfxRect(x1, 0, x2-x1, size.y), FC_WHITE);
        }
        break;
        
     case 3:
        // Quadrant
        {
            int xc = size.x / 2;
            int yc = size.y / 2;
            if (shape & 1)
                drawSolidBar(*pix, GfxRect(0, 0, xc, yc), FC_WHITE);
            if (shape & 2)
                drawSolidBar(*pix, GfxRect(xc, 0, size.x-xc, yc), FC_WHITE);
            if (shape & 4)
                drawSolidBar(*pix, GfxRect(0, yc, xc, size.y-yc), FC_WHITE);
            if (shape & 010)
                drawSolidBar(*pix, GfxRect(xc, yc, size.x-xc, size.y-yc), FC_WHITE);
        }
        break;

     case 4:
        // Shade
        {
            GfxFillPattern p;
            p[0] = p[2] = p[4] = p[6] = 0x55 * (shape & 3);
            p[1] = p[3] = p[5] = p[7] = 0x55 * ((shape >> 2) & 3);
            pix->drawBar(GfxRect(GfxPoint(0, 0), size), FC_WHITE, FC_BLACK, p, GFX_ALPHA_OPAQUE);
        }
        break;
    }
}

/** Try to synthesize a character.
    \param painter [out] draw here
    \param font    [in] for reference
    \param ch      [in] character to synthesize
    \return true if we could synthesize this character */
bool
synthesizeCharacter(RHPainter& painter, const GfxBitmapFont& font, uint32_t ch)
{
    static const uint16_t frameShapes[] = {
        000101, 000202, 001010, 002020, 030101, 030202, 031010, 032020,
        040101, 040202, 041010, 042020, 000110, 000210, 000120, 000220,
        000011, 000012, 000021, 000022, 001100, 001200, 002100, 002200,
        001001, 001002, 002001, 002002, 001110, 001210, 002110, 001120,
        002120, 002210, 001220, 002220, 001011, 001012, 002011, 001021,
        002021, 002012, 001022, 002022, 000111, 000112, 000211, 000212,
        000121, 000122, 000221, 000222, 001101, 001102, 001201, 001202,
        002101, 002102, 002201, 002202, 001111, 001112, 001211, 001212,
        002111, 001121, 002121, 002112, 002211, 001122, 001221, 002212,
        001222, 002122, 002221, 002222, 020101, 020202, 021010, 022020,
        000303, 003030, 000310, 000130, 000330, 000013, 000031, 000033,
        001300, 003100, 003300, 001003, 003001, 003003, 001310, 003130,
        003330, 001013, 003031, 003033, 000313, 000131, 000333, 001303,
        003101, 003303, 001313, 003131, 003333, 000000, 000000, 000000,
        000000, 000000, 000000, 000000, 000001, 001000, 000100, 000010,
        000002, 002000, 000200, 000020, 000201, 001020, 000102, 002010,
    };

    static const uint16_t blockShapes[] = {
        0103, 0170, 0161, 0152, 0143, 0134, 0125, 0116, 
        0107, 0206, 0205, 0204, 0203, 0202, 0201, 0200, 
        0243, 0401, 0411, 0407, 0100, 0270, 0304, 0310,
        0301, 0315, 0311, 0307, 0313, 0302, 0306, 0316,
    };

    /* We need the block glyph which determines the size of all pseudo-graphic characters */
    const GfxBitmapGlyph* blockGlyph = font.getGlyph(0x2588);
    if (blockGlyph != 0) {
        if (ch >= 0x2500 && ch < 0x2500 + countof(frameShapes) && frameShapes[ch - 0x2500] != 0) {
            painter.setSize(GfxPoint(blockGlyph->getWidth(), blockGlyph->getHeight()));
            synthesizeFrameCharacter(painter, frameShapes[ch - 0x2500]);
            painter.drawWidget();
            return true;
        }
        if (ch >= 0x2580 && ch < 0x2580 + countof(blockShapes)) {
            painter.setSize(GfxPoint(blockGlyph->getWidth(), blockGlyph->getHeight()));
            synthesizeBlockCharacter(painter, blockShapes[ch - 0x2580]);
            painter.drawWidget();
            return true;
        }
    }

    return false;
}
